import copy
import datetime
import io
import json
import os
import socket
import sys

import mock
import pytest

from uaclient import exceptions, lock, messages, status, util
from uaclient.cli.status import status_command
from uaclient.conftest import FakeNotice
from uaclient.event_logger import EventLoggerMode
from uaclient.files.notices import Notice, NoticesManager
from uaclient.files.user_config_file import UserConfigData
from uaclient.yaml import safe_load

M_PATH = "uaclient.cli.status."


RESPONSE_AVAILABLE_SERVICES = [
    {"name": "livepatch", "available": True},
    {"name": "fips", "available": False},
    {"name": "esm-infra", "available": True},
    {"name": "esm-apps", "available": True},
    {"name": "fips-updates", "available": False},
    {"name": "realtime-kernel", "available": False},
    {"name": "ros", "available": False},
    {"name": "ros-updates", "available": False},
]

RESPONSE_CONTRACT_INFO = {
    "accountInfo": {
        "createdAt": util.parse_rfc3339_date("2019-06-14T06:45:50Z"),
        "id": "some_id",
        "name": "Name",
        "type": "paid",
    },
    "contractInfo": {
        "createdAt": util.parse_rfc3339_date("2021-05-21T20:00:53Z"),
        "createdBy": "someone",
        "effectiveTo": util.parse_rfc3339_date("9999-12-31T00:00:00Z"),
        "id": "some_id",
        "name": "Name",
        "products": ["uai-essential-virtual"],
        "resourceEntitlements": [
            {
                "type": "esm-infra",
                "entitled": True,
                "obligations": {"enableByDefault": True},
            },
            {
                "type": "esm-apps",
                "entitled": False,
                "obligations": {"enableByDefault": True},
            },
            {
                "type": "livepatch",
                "entitled": True,
                "obligations": {"enableByDefault": False},
            },
            {
                "type": "support",
                "entitled": True,
                "affordances": {"supportLevel": "essential"},
            },
        ],
    },
}

SIMULATED_STATUS_ALL = """\
SERVICE          AVAILABLE  ENTITLED   AUTO_ENABLED    DESCRIPTION
esm-apps         yes        no         yes             Expanded Security Maintenance for Applications
esm-infra        yes        yes        yes             Expanded Security Maintenance for Infrastructure
fips             no         no         no              NIST-certified FIPS crypto packages
fips-updates     no         no         no              FIPS compliant crypto packages with stable security updates
livepatch        yes        yes        no              Canonical Livepatch service
realtime-kernel  no         no         no              Ubuntu kernel with PREEMPT_RT patches integrated
ros              no         no         no              Security Updates for the Robot Operating System
ros-updates      no         no         no              All Updates for the Robot Operating System
"""  # noqa: E501

SIMULATED_STATUS = """\
SERVICE          AVAILABLE  ENTITLED   AUTO_ENABLED    DESCRIPTION
esm-apps         yes        no         yes             Expanded Security Maintenance for Applications
esm-infra        yes        yes        yes             Expanded Security Maintenance for Infrastructure
livepatch        yes        yes        no              Canonical Livepatch service
"""  # noqa: E501

UNATTACHED_STATUS_ALL = """\
SERVICE          AVAILABLE  DESCRIPTION
esm-apps         yes        Expanded Security Maintenance for Applications
esm-infra        yes        Expanded Security Maintenance for Infrastructure
fips             no         NIST-certified FIPS crypto packages
fips-updates     no         FIPS compliant crypto packages with stable security updates
livepatch        yes        Canonical Livepatch service
realtime-kernel  no         Ubuntu kernel with PREEMPT_RT patches integrated
ros              no         Security Updates for the Robot Operating System
ros-updates      no         All Updates for the Robot Operating System

This machine is not attached to an Ubuntu Pro subscription.
See https://ubuntu.com/pro
"""  # noqa: E501

UNATTACHED_STATUS = """\
SERVICE          AVAILABLE  DESCRIPTION
esm-apps         yes        Expanded Security Maintenance for Applications
esm-infra        yes        Expanded Security Maintenance for Infrastructure
livepatch        yes        Canonical Livepatch service

For a list of all Ubuntu Pro services, run 'pro status --all'

This machine is not attached to an Ubuntu Pro subscription.
See https://ubuntu.com/pro
"""  # noqa: E501

ATTACHED_STATUS_ALL = """\
SERVICE          ENTITLED  STATUS       DESCRIPTION
esm-apps         no        {dash}            Expanded Security Maintenance for Applications
esm-infra        no        {dash}            Expanded Security Maintenance for Infrastructure
fips             no        {dash}            NIST-certified FIPS crypto packages
fips-updates     no        {dash}            FIPS compliant crypto packages with stable security updates
livepatch        no        {dash}            Canonical Livepatch service
realtime-kernel  no        {dash}            Ubuntu kernel with PREEMPT_RT patches integrated
ros              no        {dash}            Security Updates for the Robot Operating System
ros-updates      no        {dash}            All Updates for the Robot Operating System
{notices}{features}
Enable services with: pro enable <service>

                Account: test
           Subscription: test_contract
            Valid until: formatteddate
Technical support level: n/a
"""  # noqa: E501

ATTACHED_STATUS = """\
SERVICE          ENTITLED  STATUS       DESCRIPTION
esm-apps         no        {dash}            Expanded Security Maintenance for Applications
esm-infra        no        {dash}            Expanded Security Maintenance for Infrastructure
livepatch        no        {dash}            Canonical Livepatch service
{notices}{features}
For a list of all Ubuntu Pro services, run 'pro status --all'
Enable services with: pro enable <service>

                Account: test
           Subscription: test_contract
            Valid until: formatteddate
Technical support level: n/a
"""  # noqa: E501


SERVICES_JSON_ALL = [
    {
        "description": "Expanded Security Maintenance for Applications",
        "description_override": None,
        "entitled": "no",
        "name": "esm-apps",
        "status": "—",
        "status_details": "",
        "available": "yes",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "Expanded Security Maintenance for Infrastructure",
        "description_override": None,
        "entitled": "no",
        "name": "esm-infra",
        "status": "—",
        "status_details": "",
        "available": "yes",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "NIST-certified FIPS crypto packages",
        "description_override": None,
        "entitled": "no",
        "name": "fips",
        "status": "—",
        "status_details": "",
        "available": "no",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": (
            "FIPS compliant crypto packages with stable security updates"  # noqa
        ),
        "description_override": None,
        "entitled": "no",
        "name": "fips-updates",
        "status": "—",
        "status_details": "",
        "available": "no",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "Canonical Livepatch service",
        "description_override": None,
        "entitled": "no",
        "name": "livepatch",
        "status": "—",
        "status_details": "",
        "available": "yes",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "Ubuntu kernel with PREEMPT_RT patches integrated",
        "description_override": None,
        "entitled": "no",
        "name": "realtime-kernel",
        "status": "—",
        "status_details": "",
        "available": "no",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "Security Updates for the Robot Operating System",
        "description_override": None,
        "entitled": "no",
        "name": "ros",
        "status": "—",
        "status_details": "",
        "available": "no",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "All Updates for the Robot Operating System",
        "description_override": None,
        "entitled": "no",
        "name": "ros-updates",
        "status": "—",
        "status_details": "",
        "available": "no",
        "blocked_by": [],
        "warning": None,
    },
]

SERVICES_JSON = [
    {
        "description": "Expanded Security Maintenance for Applications",
        "description_override": None,
        "entitled": "no",
        "name": "esm-apps",
        "status": "—",
        "status_details": "",
        "available": "yes",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "Expanded Security Maintenance for Infrastructure",
        "description_override": None,
        "entitled": "no",
        "name": "esm-infra",
        "status": "—",
        "status_details": "",
        "available": "yes",
        "blocked_by": [],
        "warning": None,
    },
    {
        "description": "Canonical Livepatch service",
        "description_override": None,
        "entitled": "no",
        "name": "livepatch",
        "status": "—",
        "status_details": "",
        "available": "yes",
        "blocked_by": [],
        "warning": None,
    },
]


@mock.patch("uaclient.livepatch.on_supported_kernel", return_value=None)
@mock.patch("uaclient.system.should_reboot", return_value=False)
@mock.patch(
    "uaclient.status.get_available_resources",
    return_value=RESPONSE_AVAILABLE_SERVICES,
)
@mock.patch(
    "uaclient.status.get_contract_information",
    return_value=RESPONSE_CONTRACT_INFO,
)
@mock.patch(
    "uaclient.files.user_config_file.UserConfigFileObject.public_config",
    new_callable=mock.PropertyMock,
    return_value=UserConfigData(),
)
class TestActionStatus:
    @pytest.mark.parametrize("use_all", (True, False))
    @pytest.mark.parametrize(
        "notices,notice_status",
        (
            ([], ""),
            (
                [
                    [FakeNotice.reboot_required, "adesc"],
                    [FakeNotice.enable_reboot_required, "bdesc"],
                ],
                "\nNOTICES\nadesc\nbdesc\n",
            ),
        ),
    )
    @pytest.mark.parametrize(
        "features,feature_status",
        (
            ({}, ""),
            (
                {"one": True, "other": False, "some": "thing"},
                "\nFEATURES\none: True\nother: False\nsome: thing\n",
            ),
        ),
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.status.format_expires", return_value="formatteddate")
    def test_attached(
        self,
        _m_format_expires,
        _m_status_cache_file,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        features,
        feature_status,
        notices,
        notice_status,
        use_all,
        capsys,
        fake_machine_token_file,
    ):
        """Check that root and non-root will emit attached status"""
        fake_machine_token_file.attached = True
        mock_notice = NoticesManager()
        for notice in notices:
            mock_notice.add(notice[0], notice[1])
        with mock.patch(
            "uaclient.config.UAConfig.features",
            new_callable=mock.PropertyMock,
            return_value=features,
        ):
            assert 0 == status_command.action(
                mock.MagicMock(all=use_all, simulate_with_token=None), cfg=None
            )
        # capsys already converts colorized non-printable chars to space
        # Strip non-printables from output
        printable_stdout = capsys.readouterr()[0].replace(" " * 17, " " * 8)

        # On older versions of pytest, capsys doesn't set sys.stdout.encoding
        # to something that Python parses as UTF-8 compatible, so we get the
        # ASCII dash; testing for the "wrong" dash here is OK, because we have
        # a specific test that the correct one is used in
        # test_unicode_dash_replacement_when_unprintable
        expected_dash = "-"
        status_tmpl = ATTACHED_STATUS_ALL if use_all else ATTACHED_STATUS

        if sys.stdout.encoding and "UTF-8" in sys.stdout.encoding.upper():
            expected_dash = "\u2014"
        assert (
            status_tmpl.format(
                dash=expected_dash,
                notices=notice_status,
                features=feature_status,
            )
            == printable_stdout
        )

    @pytest.mark.parametrize(
        "use_all",
        (
            (True),
            (False),
        ),
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    def test_unattached(
        self,
        _m_status_cache_file,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        use_all,
        capsys,
        FakeConfig,
    ):
        """Check that unattached status is emitted to console"""
        cfg = FakeConfig()

        expected = UNATTACHED_STATUS_ALL if use_all else UNATTACHED_STATUS
        assert 0 == status_command.action(
            mock.MagicMock(all=use_all, simulate_with_token=None), cfg=cfg
        )
        assert expected == capsys.readouterr()[0]

    @pytest.mark.parametrize(
        "use_all",
        (
            (True),
            (False),
        ),
    )
    def test_simulated(
        self,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        use_all,
        capsys,
        FakeConfig,
    ):
        """Check that a simulated status is emitted to console"""
        cfg = FakeConfig()
        expected = SIMULATED_STATUS_ALL if use_all else SIMULATED_STATUS

        assert 0 == status_command.action(
            mock.MagicMock(all=use_all, simulate_with_token="some_token"),
            cfg=cfg,
        )
        assert expected == capsys.readouterr()[0]

    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.lock.check_lock_info")
    @mock.patch("uaclient.version.get_version", return_value="test_version")
    @mock.patch("uaclient.system.subp")
    @mock.patch(M_PATH + "time.sleep")
    def test_wait_blocks_until_lock_released(
        self,
        m_sleep,
        _m_subp,
        _m_get_version,
        m_check_lock_info,
        _m_status_cache_file,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        capsys,
        FakeConfig,
    ):
        """Check that --wait will will block and poll until lock released."""
        cfg = FakeConfig()
        mock_notice = NoticesManager()
        m_check_lock_info.return_value = (123, "pro disable")

        def fake_sleep(seconds):
            if m_sleep.call_count == 3:
                m_check_lock_info.return_value = (-1, "")
                mock_notice.remove(Notice.OPERATION_IN_PROGRESS)

        m_sleep.side_effect = fake_sleep

        with mock.patch.object(lock, "lock_data_file"):
            assert 0 == status_command.action(
                mock.MagicMock(all=False, simulate_with_token=None), cfg=cfg
            )

        assert [mock.call(1)] * 3 == m_sleep.call_args_list
        assert "...\n" + UNATTACHED_STATUS == capsys.readouterr()[0]

    @pytest.mark.parametrize(
        "use_all",
        (
            (True),
            (False),
        ),
    )
    @pytest.mark.parametrize(
        "format_type,event_logger_mode",
        (("json", EventLoggerMode.JSON), ("yaml", EventLoggerMode.YAML)),
    )
    @pytest.mark.parametrize(
        "environ",
        (
            {},
            {
                "UA_DATA_DIR": "data_dir",
                "UA_TEST": "test",
                "UA_FEATURES_ALLOW_BETA": "true",
                "UA_CONFIG_FILE": "config_file",
            },
        ),
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    def test_unattached_formats(
        self,
        _m_status_cache_file,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        environ,
        format_type,
        event_logger_mode,
        use_all,
        capsys,
        FakeConfig,
        event,
    ):
        """Check that unattached status json output is emitted to console"""
        cfg = FakeConfig()

        args = mock.MagicMock(
            format=format_type, all=use_all, simulate_with_token=None
        )
        with mock.patch.object(os, "environ", environ):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger_mode
            ), mock.patch.object(event, "_command", "status"):
                assert 0 == status_command.action(args, cfg=cfg)

        expected_environment = []
        if environ:
            expected_environment = [
                {"name": "UA_CONFIG_FILE", "value": "config_file"},
                {"name": "UA_DATA_DIR", "value": "data_dir"},
                {"name": "UA_FEATURES_ALLOW_BETA", "value": "true"},
            ]

        services = SERVICES_JSON_ALL if use_all else SERVICES_JSON
        services = [
            {
                "name": service["name"],
                "description": service["description"],
                "description_override": service["description_override"],
                "available": service["available"],
            }
            for service in services
        ]

        expected = {
            "_doc": (
                "Content provided in json response is currently "
                "considered Experimental and may change"
            ),
            "_schema_version": "0.1",
            "version": mock.ANY,
            "execution_status": status.UserFacingConfigStatus.INACTIVE.value,
            "execution_details": messages.NO_ACTIVE_OPERATIONS,
            "attached": False,
            "machine_id": None,
            "effective": None,
            "expires": None,
            "features": {},
            "notices": [],
            "services": services,
            "environment_vars": expected_environment,
            "contract": {
                "id": "",
                "name": "",
                "created_at": "",
                "products": [],
                "tech_support_level": "n/a",
            },
            "account": {
                "name": "",
                "id": "",
                "created_at": "",
                "external_account_ids": [],
            },
            "config_path": None,
            "config": {
                "data_dir": mock.ANY,
                "ua_config": mock.ANY,
                "log_file": mock.ANY,
            },
            "simulated": False,
            "errors": [],
            "warnings": [],
            "result": "success",
        }

        if format_type == "json":
            assert expected == json.loads(capsys.readouterr()[0])
        else:
            assert expected == safe_load(capsys.readouterr()[0])

    @pytest.mark.parametrize(
        "format_type,event_logger_mode",
        (("json", EventLoggerMode.JSON), ("yaml", EventLoggerMode.YAML)),
    )
    @pytest.mark.parametrize(
        "environ",
        (
            {},
            {
                "UA_DATA_DIR": "data_dir",
                "UA_TEST": "test",
                "UA_FEATURES_ALLOW_BETA": "true",
                "UA_CONFIG_FILE": "config_file",
            },
        ),
    )
    @pytest.mark.parametrize("use_all", (True, False))
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    def test_attached_formats(
        self,
        _m_status_cache_file,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        use_all,
        environ,
        format_type,
        event_logger_mode,
        capsys,
        FakeConfig,
        fake_machine_token_file,
        event,
    ):
        """Check that unattached status json output is emitted to console"""
        fake_machine_token_file.attached = True

        args = mock.MagicMock(
            format=format_type, all=use_all, simulate_with_token=None
        )

        with mock.patch.object(os, "environ", environ):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger_mode
            ), mock.patch.object(event, "_command", "status"):
                with mock.patch(
                    "uaclient.status._get_blocked_by_services", return_value=[]
                ):
                    assert 0 == status_command.action(args, cfg=FakeConfig())

        expected_environment = []
        if environ:
            expected_environment = [
                {"name": "UA_CONFIG_FILE", "value": "config_file"},
                {"name": "UA_DATA_DIR", "value": "data_dir"},
                {"name": "UA_FEATURES_ALLOW_BETA", "value": "true"},
            ]

        services = SERVICES_JSON_ALL if use_all else SERVICES_JSON

        if format_type == "json":
            contract_created_at = "2020-05-08T19:02:26+00:00"
            account_created_at = "2019-06-14T06:45:50+00:00"
            expires = "2040-05-08T19:02:26+00:00"
            effective = "2000-05-08T19:02:26+00:00"
        else:
            contract_created_at = datetime.datetime(
                2020, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
            )
            account_created_at = datetime.datetime(
                2019, 6, 14, 6, 45, 50, tzinfo=datetime.timezone.utc
            )
            expires = datetime.datetime(
                2040, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
            )
            effective = datetime.datetime(
                2000, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
            )

        tech_support_level = status.UserFacingStatus.INAPPLICABLE.value
        expected = {
            "_doc": (
                "Content provided in json response is currently "
                "considered Experimental and may change"
            ),
            "_schema_version": "0.1",
            "version": mock.ANY,
            "execution_status": status.UserFacingConfigStatus.INACTIVE.value,
            "execution_details": messages.NO_ACTIVE_OPERATIONS,
            "attached": True,
            "machine_id": "test_machine_id",
            "effective": effective,
            "expires": expires,
            "features": {},
            "notices": [],
            "services": services,
            "environment_vars": expected_environment,
            "contract": {
                "id": "cid",
                "name": "test_contract",
                "created_at": contract_created_at,
                "products": ["free"],
                "tech_support_level": tech_support_level,
            },
            "account": {
                "id": "acct-1",
                "name": "test",
                "created_at": account_created_at,
                "external_account_ids": [{"IDs": ["id1"], "origin": "AWS"}],
            },
            "config_path": None,
            "config": {
                "data_dir": mock.ANY,
                "ua_config": mock.ANY,
                "log_file": mock.ANY,
            },
            "simulated": False,
            "errors": [],
            "warnings": [],
            "result": "success",
        }

        if format_type == "json":
            assert expected == json.loads(capsys.readouterr()[0])
        else:
            yaml_output = safe_load(capsys.readouterr()[0])

            # On earlier versions of pyyaml, we don't add the timezone
            # info when converting a date string into a datetime object.
            # Since we only want to test if we are producing a valid
            # yaml file in the status output, we can manually add
            # the timezone info to make the test work as expected
            for key, value in yaml_output.items():
                if isinstance(value, datetime.datetime):
                    yaml_output[key] = value.replace(
                        tzinfo=datetime.timezone.utc
                    )

                elif isinstance(value, dict):
                    for inner_key, inner_value in value.items():
                        if isinstance(inner_value, datetime.datetime):
                            yaml_output[key][inner_key] = inner_value.replace(
                                tzinfo=datetime.timezone.utc
                            )

            assert expected == yaml_output

    @pytest.mark.parametrize(
        "format_type,event_logger_mode",
        (("json", EventLoggerMode.JSON), ("yaml", EventLoggerMode.YAML)),
    )
    @pytest.mark.parametrize("use_all", (True, False))
    def test_simulated_formats(
        self,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        use_all,
        format_type,
        event_logger_mode,
        capsys,
        FakeConfig,
        event,
    ):
        """Check that simulated status json output is emitted to console"""
        cfg = FakeConfig()

        args = mock.MagicMock(
            format=format_type, all=use_all, simulate_with_token="some_token"
        )

        with mock.patch.object(
            event, "_event_logger_mode", event_logger_mode
        ), mock.patch.object(event, "_command", "status"):
            assert 0 == status_command.action(args, cfg=cfg)

        services = [
            {
                "auto_enabled": "yes",
                "available": "yes",
                "description": "Expanded Security Maintenance for Applications",  # noqa
                "entitled": "no",
                "name": "esm-apps",
            },
            {
                "auto_enabled": "yes",
                "available": "yes",
                "description": "Expanded Security Maintenance for Infrastructure",  # noqa
                "entitled": "yes",
                "name": "esm-infra",
            },
            {
                "auto_enabled": "no",
                "available": "no",
                "description": "NIST-certified FIPS crypto packages",
                "entitled": "no",
                "name": "fips",
            },
            {
                "auto_enabled": "no",
                "available": "no",
                "description": "FIPS compliant crypto packages "
                "with stable security updates",
                "entitled": "no",
                "name": "fips-updates",
            },
            {
                "auto_enabled": "no",
                "available": "yes",
                "description": "Canonical Livepatch service",
                "entitled": "yes",
                "name": "livepatch",
            },
            {
                "auto_enabled": "no",
                "available": "no",
                "description": "Ubuntu kernel with PREEMPT_RT patches"
                " integrated",
                "entitled": "no",
                "name": "realtime-kernel",
            },
            {
                "auto_enabled": "no",
                "available": "no",
                "description": (
                    "Security Updates for the Robot Operating System"
                ),
                "entitled": "no",
                "name": "ros",
            },
            {
                "auto_enabled": "no",
                "available": "no",
                "description": "All Updates for the Robot Operating System",
                "entitled": "no",
                "name": "ros-updates",
            },
        ]

        expected_services = sorted(services, key=lambda x: x["name"])
        if not use_all:
            expected_services = [
                service
                for service in services
                if service["available"] == "yes"
            ]

        expected = {
            "_doc": "Content provided in json response is currently considered"
            " Experimental and may change",
            "_schema_version": "0.1",
            "attached": False,
            "machine_id": None,
            "features": {},
            "notices": [],
            "account": {
                "created_at": util.parse_rfc3339_date("2019-06-14T06:45:50Z"),
                "external_account_ids": [],
                "id": "some_id",
                "name": "Name",
            },
            "contract": {
                "created_at": util.parse_rfc3339_date("2021-05-21T20:00:53Z"),
                "id": "some_id",
                "name": "Name",
                "products": ["uai-essential-virtual"],
                "tech_support_level": "essential",
            },
            "environment_vars": [],
            "execution_status": "inactive",
            "execution_details": "No Ubuntu Pro operations are running",
            "expires": util.parse_rfc3339_date("9999-12-31T00:00:00Z"),
            "effective": None,
            "services": expected_services,
            "simulated": True,
            "version": mock.ANY,
            "config_path": None,
            "config": {
                "data_dir": mock.ANY,
                "ua_config": mock.ANY,
                "log_file": mock.ANY,
            },
            "errors": [],
            "warnings": [],
            "result": "success",
        }

        if format_type == "json":
            assert expected == json.loads(
                capsys.readouterr()[0], cls=util.DatetimeAwareJSONDecoder
            )
        else:
            expected["account"]["created_at"] = safe_load(
                "2019-06-14 06:45:50+00:00"
            )
            expected["contract"]["created_at"] = safe_load(
                "2021-05-21 20:00:53+00:00"
            )
            expected["expires"] = safe_load("9999-12-31 00:00:00+00:00")
            assert expected == safe_load(capsys.readouterr()[0])

    def test_error_on_connectivity_errors(
        self,
        _m_public_config,
        _m_get_contract_information,
        m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        FakeConfig,
    ):
        """Raise ConnectivityError on connectivity issues"""
        m_get_avail_resources.side_effect = exceptions.ConnectivityError(
            cause=socket.gaierror(-2, "Name or service not known"), url="url"
        )

        cfg = FakeConfig()

        with pytest.raises(exceptions.ConnectivityError):
            status_command.action(
                mock.MagicMock(all=False, simulate_with_token=None), cfg=cfg
            )

    @pytest.mark.parametrize("use_all", (True, False))
    @pytest.mark.parametrize(
        "encoding,expected_dash",
        (("utf-8", "\u2014"), ("UTF-8", "\u2014"), ("ascii", "-")),
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.status.format_expires", return_value="formatteddate")
    def test_unicode_dash_replacement_when_unprintable(
        self,
        _m_format_expires,
        _m_status_cache_file,
        _m_public_config,
        _m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        encoding,
        expected_dash,
        use_all,
        FakeConfig,
        fake_machine_token_file,
    ):
        # This test can't use capsys because it doesn't emulate sys.stdout
        # encoding accurately in older versions of pytest
        underlying_stdout = io.BytesIO()
        fake_stdout = io.TextIOWrapper(underlying_stdout, encoding=encoding)
        fake_machine_token_file.attached = True

        with mock.patch("sys.stdout", fake_stdout):
            status_command.action(
                mock.MagicMock(all=use_all, simulate_with_token=None),
                cfg=FakeConfig(),
            )

        fake_stdout.flush()  # Make sure all output is in underlying_stdout
        out = underlying_stdout.getvalue().decode(encoding)

        # Colour codes are converted to spaces, so strip them out for
        # comparison
        out = out.replace(" " * 17, " " * 8)

        if not use_all:
            expected_out = ATTACHED_STATUS.format(
                dash=expected_dash, notices="", features=""
            )
        else:
            expected_out = ATTACHED_STATUS_ALL.format(
                dash=expected_dash, notices="", features=""
            )

        assert expected_out == out

    @pytest.mark.parametrize(
        "exception_to_throw,exception_type,exception_message",
        (
            (
                exceptions.ConnectivityError(Exception("Not found"), "url"),
                exceptions.ConnectivityError,
                "Not found",
            ),
            (
                exceptions.ContractAPIError(
                    url="url",
                    code=401,
                    body=json.dumps({"message": "unauthorized"}),
                ),
                exceptions.AttachInvalidTokenError,
                "Invalid token. See https://ubuntu.com/pro",
            ),
        ),
    )
    def test_errors_are_raised_appropriately(
        self,
        _m_public_config,
        m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        exception_to_throw,
        exception_type,
        exception_message,
        capsys,
        FakeConfig,
    ):
        """Check that simulated status json/yaml output raises errors."""

        m_get_contract_information.side_effect = exception_to_throw

        cfg = FakeConfig()

        args = mock.MagicMock(
            format="json", all=False, simulate_with_token="some_token"
        )

        with pytest.raises(exception_type) as exc:
            status_command.action(args, cfg=cfg)

        assert exc.type == exception_type
        assert exception_message in getattr(exc.value, "msg", exc.value.args)

    @pytest.mark.parametrize(
        "token_to_use,warning_message,contract_field,date_value",
        (
            (
                "expired_token",
                (
                    "Attach denied:\n"
                    'Contract "some_id" expired on December 31, 2019\n'
                    "Visit https://ubuntu.com/pro/dashboard to manage "
                    "contract tokens."
                ),
                "effectiveTo",
                util.parse_rfc3339_date("2019-12-31T00:00:00Z"),
            ),
            (
                "token_not_valid_yet",
                (
                    "Attach denied:\n"
                    'Contract "some_id" is not effective until December 31, '
                    "9999\n"
                    "Visit https://ubuntu.com/pro/dashboard to manage "
                    "contract tokens."
                ),
                "effectiveFrom",
                util.parse_rfc3339_date("9999-12-31T00:00:00Z"),
            ),
        ),
    )
    @pytest.mark.parametrize(
        "format_type,event_logger_mode",
        (("json", EventLoggerMode.JSON), ("yaml", EventLoggerMode.YAML)),
    )
    def test_errors_for_token_dates(
        self,
        _m_public_config,
        m_get_contract_information,
        _m_get_avail_resources,
        _m_should_reboot,
        _m_on_supported_kernel,
        format_type,
        event_logger_mode,
        token_to_use,
        warning_message,
        contract_field,
        date_value,
        capsys,
        FakeConfig,
        event,
    ):
        """Check errors for expired tokens, and not valid yet tokens."""

        def contract_info_side_effect(cfg, token):
            response = copy.deepcopy(RESPONSE_CONTRACT_INFO)
            response["contractInfo"][contract_field] = date_value
            return response

        m_get_contract_information.side_effect = contract_info_side_effect

        cfg = FakeConfig()
        args = mock.MagicMock(
            format=format_type, all=False, simulate_with_token=token_to_use
        )

        with mock.patch.object(
            event, "_event_logger_mode", event_logger_mode
        ), mock.patch.object(event, "_command", "status"):
            assert 1 == status_command.action(args, cfg=cfg)

        if format_type == "json":
            output = json.loads(capsys.readouterr()[0])
        else:
            output = safe_load(capsys.readouterr()[0])

        assert output["errors"][0]["message"] == warning_message



"""Tests related to uaclient.util module."""

import datetime
import json

import mock
import pytest

from uaclient import exceptions, messages, secret_manager, util


@pytest.fixture
def secret_manager_fixture():
    manager = secret_manager.SecretManager()
    secrets = [
        "SEKRET",
        "S3-Kr3T",
        "S3kR3T",
        "SEket.124-_ys",
        "reg key",
        "reg-key",
    ]
    for secret in secrets:
        manager.add_secret(secret)
    return manager


class TestGetDictDeltas:
    @pytest.mark.parametrize(
        "value1,value2", (("val1", "val2"), ([1], [2]), ((1, 2), (3, 4)))
    )
    def test_non_dict_diffs_return_new_value(self, value1, value2):
        """When two values differ and are not a dict return the new value."""
        expected = {"key": value2}
        assert expected == util.get_dict_deltas(
            {"key": value1}, {"key": value2}
        )

    def test_diffs_return_new_keys_and_values(self):
        """New keys previously absent will be returned in the delta."""
        expected = {"newkey": "val"}
        assert expected == util.get_dict_deltas(
            {"k": "v"}, {"newkey": "val", "k": "v"}
        )

    def test_diffs_return_dropped_keys_set_dropped(self):
        """Old keys which are now dropped are returned as DROPPED_KEY."""
        expected = {"oldkey": util.DROPPED_KEY, "oldkey2": util.DROPPED_KEY}
        assert expected == util.get_dict_deltas(
            {"oldkey": "v", "k": "v", "oldkey2": {}}, {"k": "v"}
        )

    def test_return_only_keys_which_represent_deltas(self):
        """Only return specific keys which have deltas."""
        orig_dict = {
            "1": "1",
            "2": "orig2",
            "3": {"3.1": "3.1", "3.2": "orig3.2"},
            "4": {"4.1": "4.1"},
        }
        new_dict = {
            "1": "1",
            "2": "new2",
            "3": {"3.1": "3.1", "3.2": "new3.2"},
            "4": {"4.1": "4.1"},
        }
        expected = {"2": "new2", "3": {"3.2": "new3.2"}}
        assert expected == util.get_dict_deltas(orig_dict, new_dict)


JSON_TEST_PAIRS = (
    ("a", '"a"'),
    (1, "1"),
    ({"a": 1}, '{"a": 1}'),
    # See the note in DatetimeAwareJSONDecoder for why this datetime is in a
    # dict
    (
        {
            "dt": datetime.datetime(
                2019, 7, 25, 14, 35, 51, tzinfo=datetime.timezone.utc
            )
        },
        '{"dt": "2019-07-25T14:35:51+00:00"}',
    ),
)


class TestDatetimeAwareJSONEncoder:
    @pytest.mark.parametrize("input,out", JSON_TEST_PAIRS)
    def test_encode(self, input, out):
        assert out == json.dumps(input, cls=util.DatetimeAwareJSONEncoder)


class TestDatetimeAwareJSONDecoder:
    # Note that the parameter names are flipped from
    # TestDatetimeAwareJSONEncoder
    @pytest.mark.parametrize("out,input", JSON_TEST_PAIRS)
    def test_encode(self, input, out):
        assert out == json.loads(input, cls=util.DatetimeAwareJSONDecoder)


@mock.patch("builtins.input")
class TestPromptForConfirmation:
    @pytest.mark.parametrize(
        "return_value,user_input",
        [(True, yes_input) for yes_input in ["y", "Y", "yes", "YES", "YeS"]]
        + [
            (False, no_input)
            for no_input in ["n", "N", "no", "NO", "No", "asdf", "", "\nfoo\n"]
        ],
    )
    def test_input_conversion(self, m_input, return_value, user_input):
        m_input.return_value = user_input
        assert return_value == util.prompt_for_confirmation()

    @pytest.mark.parametrize(
        "assume_yes,message,input_calls",
        [
            (True, "message ignored on assume_yes=True", []),
            (False, "", [mock.call("Are you sure? (y/N) ")]),
            (False, "Custom yep? (y/N) ", [mock.call("Custom yep? (y/N) ")]),
        ],
    )
    def test_prompt_text(self, m_input, assume_yes, message, input_calls):
        util.prompt_for_confirmation(msg=message, assume_yes=assume_yes)

        assert input_calls == m_input.call_args_list


class TestIsConfigValueTrue:
    @pytest.mark.parametrize(
        "config_dict, return_val",
        [
            ({}, False),
            ({}, False),
            (None, False),
            ({None}, False),
            ({"allow_beta": "true"}, True),
            ({"allow_beta": "True"}, True),
            ({"allow_beta": "false"}, False),
            ({"allow_beta": "False"}, False),
        ],
    )
    def test_is_config_value_true(self, config_dict, return_val, FakeConfig):
        cfg = FakeConfig()
        cfg.override_features(config_dict)
        actual_val = util.is_config_value_true(
            config=cfg.cfg, path_to_value="features.allow_beta"
        )
        assert return_val == actual_val

    @pytest.mark.parametrize(
        "config_dict, key_val",
        [
            ({"allow_beta": "tru"}, "tru"),
            ({"allow_beta": "Tre"}, "Tre"),
            ({"allow_beta": "flse"}, "flse"),
            ({"allow_beta": "Fale"}, "Fale"),
        ],
    )
    def test_exception_is_config_value_true(
        self, config_dict, key_val, FakeConfig
    ):
        path_to_value = "features.allow_beta"
        cfg = FakeConfig()
        cfg.override_features(config_dict)
        with pytest.raises(exceptions.InvalidBooleanConfigValue) as excinfo:
            util.is_config_value_true(
                config=cfg.cfg, path_to_value=path_to_value
            )

        expected_msg = messages.E_INVALID_BOOLEAN_CONFIG_VALUE.format(
            path_to_value=path_to_value,
            expected_value="boolean string: true or false",
            value=key_val,
        ).msg
        assert expected_msg == str(excinfo.value)


class TestRedactSensitiveLogs:
    @pytest.mark.parametrize(
        "raw_log,expected",
        (
            ("Super valuable", "Super valuable"),
            (
                "Hi 'Bearer not the droids you are looking for', data",
                "Hi 'Bearer <REDACTED>', data",
            ),
            (
                "Hi 'Bearer not the droids you are looking for', data",
                "Hi 'Bearer <REDACTED>', data",
            ),
            (
                "Executed with sys.argv: ['/usr/bin/ua', 'attach', 'SEKRET']",
                "Executed with sys.argv:"
                " ['/usr/bin/ua', 'attach', '<REDACTED>']",
            ),
            (
                "'resourceTokens': [{'token': 'SEKRET', 'type': 'cc-eal'}]'",
                "'resourceTokens':"
                " [{'token': '<REDACTED>', 'type': 'cc-eal'}]'",
            ),
            (
                "'machineToken': 'SEKRET', 'machineTokenInfo': 'blah'",
                "'machineToken': '<REDACTED>', 'machineTokenInfo': 'blah'",
            ),
            (
                "Failed running command '/usr/lib/apt/apt-helper download-file"
                "https://bearer:S3-Kr3T@esm.ubuntu.com/infra/ubuntu/pool/ "
                "[exit(100)]. Message: Download of file failed"
                " pkgAcquire::Run (13: Permission denied)",
                "Failed running command '/usr/lib/apt/apt-helper download-file"
                "https://bearer:<REDACTED>@esm.ubuntu.com/infra/ubuntu/pool/ "
                "[exit(100)]. Message: Download of file failed"
                " pkgAcquire::Run (13: Permission denied)",
            ),
            (
                "/snap/bin/canonical-livepatch enable S3-Kr3T, foobar",
                "/snap/bin/canonical-livepatch enable <REDACTED> foobar",
            ),
            (
                "Contract value for 'resourceToken' changed to S3kR3T",
                "Contract value for 'resourceToken' changed to <REDACTED>",
            ),
            (
                "data: {'contractToken': 'SEKRET', "
                "'contractTokenInfo':{'expiry'}}",
                "data: {'contractToken': '<REDACTED>', "
                "'contractTokenInfo':{'expiry'}}",
            ),
            (
                "data: {'resourceToken': 'SEKRET', "
                "'entitlement': {'affordances':'blah blah' }}",
                "data: {'resourceToken': '<REDACTED>', "
                "'entitlement': {'affordances':'blah blah' }}",
            ),
            (
                "https://contracts.canonical.com/v1/resources/livepatch"
                "?token=SEKRET: invalid token",
                "https://contracts.canonical.com/v1/resources/livepatch"
                "?token=<REDACTED> invalid token",
            ),
            (
                'data: {"identityToken": "SEket.124-_ys"}',
                'data: {"identityToken": "<REDACTED>"}',
            ),
            (
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
            ),
            (
                "response: "
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
                "response: "
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: <REDACTED>",
            ),
            (
                "'token': 'SEKRET'",
                "'token': '<REDACTED>'",
            ),
            (
                "'userCode': 'SEKRET'",
                "'userCode': '<REDACTED>'",
            ),
            (
                "'magic_token=SEKRET'",
                "'magic_token=<REDACTED>'",
            ),
            (
                "--account-name name --registration-key=reg-key --silent",
                "--account-name name --registration-key=<REDACTED> --silent",
            ),
            (
                '--account-name name --registration-key="reg key" --silent',
                "--account-name name --registration-key=<REDACTED> --silent",
            ),
            (
                "--account-name name --registration-key='reg key' --silent",
                "--account-name name --registration-key=<REDACTED> --silent",
            ),
            (
                "--account-name name --registration-key reg-key --silent",
                "--account-name name --registration-key <REDACTED> --silent",
            ),
            (
                '--account-name name --registration-key "reg key" --silent',
                "--account-name name --registration-key <REDACTED> --silent",
            ),
            (
                "--account-name name --registration-key 'reg key' --silent",
                "--account-name name --registration-key <REDACTED> --silent",
            ),
            (
                "--account-name name -p reg-key --silent",
                "--account-name name -p <REDACTED> --silent",
            ),
            (
                '--account-name name -p "reg key" --silent',
                "--account-name name -p <REDACTED> --silent",
            ),
            (
                "--account-name name -p 'reg key' --silent",
                "--account-name name -p <REDACTED> --silent",
            ),
        ),
    )
    def test_redact_all_matching_regexs(self, raw_log, expected):
        """Redact all sensitive matches from log messages."""
        assert expected == util.redact_sensitive_logs(raw_log)

    @pytest.mark.parametrize(
        "raw_log,expected",
        (
            ("Super valuable", "Super valuable"),
            (
                "Executed with sys.argv: ['/usr/bin/ua', 'attach', 'SEKRET']",
                "Executed with sys.argv:"
                " ['/usr/bin/ua', 'attach', '<REDACTED>']",
            ),
            (
                "'resourceTokens': [{'token': 'SEKRET', 'type': 'cc-eal'}]'",
                "'resourceTokens':"
                " [{'token': '<REDACTED>', 'type': 'cc-eal'}]'",
            ),
            (
                "'machineToken': 'SEKRET', 'machineTokenInfo': 'blah'",
                "'machineToken': '<REDACTED>', 'machineTokenInfo': 'blah'",
            ),
            (
                "Failed running command '/usr/lib/apt/apt-helper download-file"
                "https://bearer:S3-Kr3T@esm.ubuntu.com/infra/ubuntu/pool/ "
                "[exit(100)]. Message: Download of file failed"
                " pkgAcquire::Run (13: Permission denied)",
                "Failed running command '/usr/lib/apt/apt-helper download-file"
                "https://bearer:<REDACTED>@esm.ubuntu.com/infra/ubuntu/pool/ "
                "[exit(100)]. Message: Download of file failed"
                " pkgAcquire::Run (13: Permission denied)",
            ),
            (
                "/snap/bin/canonical-livepatch enable S3-Kr3T, foobar",
                "/snap/bin/canonical-livepatch enable <REDACTED>, foobar",
            ),
            (
                "Contract value for 'resourceToken' changed to S3kR3T",
                "Contract value for 'resourceToken' changed to <REDACTED>",
            ),
            (
                "data: {'contractToken': 'SEKRET', "
                "'contractTokenInfo':{'expiry'}}",
                "data: {'contractToken': '<REDACTED>', "
                "'contractTokenInfo':{'expiry'}}",
            ),
            (
                "data: {'resourceToken': 'SEKRET', "
                "'entitlement': {'affordances':'blah blah' }}",
                "data: {'resourceToken': '<REDACTED>', "
                "'entitlement': {'affordances':'blah blah' }}",
            ),
            (
                "https://contracts.canonical.com/v1/resources/livepatch"
                "?token=SEKRET: invalid token",
                "https://contracts.canonical.com/v1/resources/livepatch"
                "?token=<REDACTED>: invalid token",
            ),
            (
                'data: {"identityToken": "SEket.124-_ys"}',
                'data: {"identityToken": "<REDACTED>"}',
            ),
            (
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
            ),
            (
                "'token': 'SEKRET'",
                "'token': '<REDACTED>'",
            ),
            (
                "'userCode': 'SEKRET'",
                "'userCode': '<REDACTED>'",
            ),
            (
                "'magic_token=SEKRET'",
                "'magic_token=<REDACTED>'",
            ),
            (
                "--account-name name --registration-key=reg-key --silent",
                "--account-name name --registration-key=<REDACTED> --silent",
            ),
            (
                '--account-name name --registration-key="reg key" --silent',
                '--account-name name --registration-key="<REDACTED>" --silent',
            ),
            (
                "--account-name name --registration-key='reg key' --silent",
                "--account-name name --registration-key='<REDACTED>' --silent",
            ),
            (
                "--account-name name --registration-key reg-key --silent",
                "--account-name name --registration-key <REDACTED> --silent",
            ),
            (
                '--account-name name --registration-key "reg key" --silent',
                '--account-name name --registration-key "<REDACTED>" --silent',
            ),
            (
                "--account-name name --registration-key 'reg key' --silent",
                "--account-name name --registration-key '<REDACTED>' --silent",
            ),
            (
                "--account-name name -p reg-key --silent",
                "--account-name name -p <REDACTED> --silent",
            ),
            (
                '--account-name name -p "reg key" --silent',
                '--account-name name -p "<REDACTED>" --silent',
            ),
            (
                "--account-name name -p 'reg key' --silent",
                "--account-name name -p '<REDACTED>' --silent",
            ),
        ),
    )
    def test_secret_redaction_filter(
        self, secret_manager_fixture, raw_log, expected
    ):
        """Redact all sensitive matches from log messages."""
        assert expected == secret_manager_fixture.redact_secrets(raw_log)


class TestParseRFC3339Date:
    @pytest.mark.parametrize(
        "datestring,expected",
        [
            (
                "2001-02-03T04:05:06",
                datetime.datetime(
                    2001, 2, 3, 4, 5, 6, tzinfo=datetime.timezone.utc
                ),
            ),
            (
                "2001-02-03T04:05:06.123456",
                datetime.datetime(
                    2001, 2, 3, 4, 5, 6, tzinfo=datetime.timezone.utc
                ),
            ),
            (
                "2001-02-03T04:05:06Z",
                datetime.datetime(
                    2001, 2, 3, 4, 5, 6, tzinfo=datetime.timezone.utc
                ),
            ),
            (
                "2001-02-03T04:05:06-08:00",
                datetime.datetime(
                    2001,
                    2,
                    3,
                    4,
                    5,
                    6,
                    tzinfo=datetime.timezone(-datetime.timedelta(hours=8)),
                ),
            ),
            (
                "2001-02-03T04:05:06+03:00",
                datetime.datetime(
                    2001,
                    2,
                    3,
                    4,
                    5,
                    6,
                    tzinfo=datetime.timezone(datetime.timedelta(hours=3)),
                ),
            ),
            (
                "2021-05-07T09:46:37.791Z",
                datetime.datetime(
                    2021, 5, 7, 9, 46, 37, tzinfo=datetime.timezone.utc
                ),
            ),
            (
                "2021-05-28T14:42:37.944609726-04:00",
                datetime.datetime(
                    2021,
                    5,
                    28,
                    14,
                    42,
                    37,
                    tzinfo=datetime.timezone(-datetime.timedelta(hours=4)),
                ),
            ),
        ],
    )
    def test_parse_rfc3339_date_from_golang(self, datestring, expected):
        """
        Check we are able to parse dates generated by golang's MarshalJSON
        """
        assert expected == util.parse_rfc3339_date(datestring)


class TestHandleUnicodeCharacters:
    @pytest.mark.parametrize(
        "encoding", ((None), ("utf-8"), ("UTF-8"), ("test"))
    )
    @pytest.mark.parametrize(
        "message,modified_message",
        (
            (messages.OKGREEN_CHECK + " test", "test"),
            (messages.FAIL_X + " fail", "fail"),
            ("\u2014 blah", "- blah"),
            ("\xfcblah", "blah"),
        ),
    )
    def test_handle_unicode_characters(
        self, message, modified_message, encoding
    ):
        expected_message = message
        if encoding is None or encoding.upper() != "UTF-8":
            expected_message = modified_message

        with mock.patch("sys.stdout") as m_stdout:
            type(m_stdout).encoding = mock.PropertyMock(return_value=encoding)
            assert expected_message == util.handle_unicode_characters(message)


class TestGetProEnvironment:
    @mock.patch(
        "os.environ",
        {
            "UA_CONFIG_FILE": "example_config_file",
            "UA_INVALID_KEY": "some_value",
            "NOT_EVEN_UA": "some_other_value",
            "UA_FEATURES_WOW": "cool_feature",
            "UA_LOG_LEVEL": "DEBUG",
        },
    )
    def test_get_pro_environment(self):
        expected = {
            "UA_CONFIG_FILE": "example_config_file",
            "UA_FEATURES_WOW": "cool_feature",
            "UA_LOG_LEVEL": "DEBUG",
        }
        assert expected == util.get_pro_environment()


class TestDeduplicateArches:
    @pytest.mark.parametrize(
        ["arches", "expected"],
        [
            ([], []),
            (["anything"], ["anything"]),
            (["amd64"], ["amd64"]),
            (["amd64", "x86_64"], ["amd64"]),
            (
                ["amd64", "ppc64el", "ppc64le", "s390x", "x86_64"],
                ["amd64", "ppc64el", "s390x"],
            ),
            (["amd64", "i386", "i686", "x86_64"], ["amd64", "i386"]),
            (
                ["amd64", "i386", "i686", "x86_64", "armhf", "arm64"],
                ["amd64", "arm64", "armhf", "i386"],
            ),
            (
                [
                    "x86_64",
                    "amd64",
                    "i686",
                    "i386",
                    "ppc64le",
                    "aarch64",
                    "arm64",
                    "armv7l",
                    "armhf",
                    "s390x",
                ],
                ["amd64", "arm64", "armhf", "i386", "ppc64el", "s390x"],
            ),
        ],
    )
    def test_deduplicate_arches(self, arches, expected):
        assert expected == util.deduplicate_arches(arches)


class TestReplaceLoggerName:
    @pytest.mark.parametrize(
        ["logger_name", "new_logger_name"],
        (
            ("uaclient.module.name1", "ubuntupro.module.name1"),
            ("log.module1.name2", "ubuntupro.module1.name2"),
            ("prolog.lang.old", "ubuntupro.lang.old"),
            ("", ""),
            (".", "ubuntupro."),
            ("module1", "ubuntupro"),
        ),
    )
    def test_replace_top_level_logger_name(self, logger_name, new_logger_name):
        assert (
            util.replace_top_level_logger_name(logger_name) == new_logger_name
        )


class TestSetFilenameExtension:
    @pytest.mark.parametrize(
        "input_string,extension,output_string",
        (
            ("virus.exe", "test", "virus.test"),
            (
                "/many/dots.and.slashes/in/this.file",
                "test",
                "/many/dots.and.slashes/in/this.test",
            ),
            ("/no/change/when.same", "same", "/no/change/when.same"),
            ("no_extension", "test", "no_extension.test"),
            (
                "/with.previous.dots/no_extension",
                "test",
                "/with.previous.dots/no_extension.test",
            ),
            ("d.o.t.s", "test", "d.o.t.test"),
            (".dotfile", "test", ".dotfile.test"),
        ),
    )
    def test_set_filename_extension(
        self, input_string, extension, output_string
    ):
        assert (
            util.set_filename_extension(input_string, extension)
            == output_string
        )


def success_call(**kwargs):
    if kwargs.get("assume_yes"):
        print("success")
        return True
    return False


def fail_call(a=None, **kwargs):
    print("fail %s" % a)
    return False


class TestHandleMessageOperations:
    @pytest.mark.parametrize(
        "msg_ops, retval, output",
        (
            ([], True, ""),
            (["msg1", "msg2"], True, "msg1\nmsg2\n"),
            (
                [(success_call, {}), "msg1", (fail_call, {"a": 1}), "msg2"],
                False,
                "success\nmsg1\nfail 1\n",
            ),
        ),
    )
    def test_handle_message_operations_for_strings_and_callables(
        self, msg_ops, retval, output, capsys
    ):
        assert retval is util.handle_message_operations(
            msg_ops, assume_yes=True
        )
        out, _err = capsys.readouterr()
        assert output == out



import itertools
import re
import sys
import threading
import time
from functools import wraps
from typing import Optional

from uaclient import (
    actions,
    api,
    daemon,
    entitlements,
    event_logger,
    exceptions,
    lock,
    messages,
    status,
    util,
)
from uaclient.api.u.pro.security.fix._common import CVE_OR_USN_REGEX
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.apt import AptProxyScope, setup_apt_proxy
from uaclient.config import UAConfig
from uaclient.files import machine_token

event = event_logger.get_event_logger()


CLEAR_LINE_ANSI_CODE = "\033[K"


class CLIEnableDisableProgress(api.AbstractProgress):
    def __init__(self, *, assume_yes: bool):
        self.is_interactive = not assume_yes
        self.assume_yes = assume_yes

    def progress(
        self,
        *,
        total_steps: int,
        done_steps: int,
        previous_step_message: Optional[str],
        current_step_message: Optional[str]
    ):
        if current_step_message is not None:
            print(current_step_message)

    def _on_event(self, event: str, payload):
        if event == "info":
            print(payload)
        elif event == "message_operation":
            if not util.handle_message_operations(payload, self.assume_yes):
                raise exceptions.PromptDeniedError()


def _null_print(*args, **kwargs):
    pass


def create_interactive_only_print_function(json_output: bool):
    if json_output:
        return _null_print
    else:
        return print


def assert_lock_file(lock_holder=None):
    """Decorator asserting exclusive access to lock file"""

    def wrapper(f):
        @wraps(f)
        def new_f(*args, cfg, **kwargs):
            with lock.RetryLock(lock_holder=lock_holder, sleep_time=1):
                retval = f(*args, cfg=cfg, **kwargs)
            return retval

        return new_f

    return wrapper


def assert_root(f):
    """Decorator asserting root user"""

    @wraps(f)
    def new_f(*args, **kwargs):
        if not util.we_are_currently_root():
            raise exceptions.NonRootUserError()
        else:
            return f(*args, **kwargs)

    return new_f


def verify_json_format_args(f):
    """Decorator to verify if correct params are used for json format"""

    @wraps(f)
    def new_f(cmd_args, *args, **kwargs):
        if not cmd_args:
            return f(cmd_args, *args, **kwargs)

        if cmd_args.format == "json" and not cmd_args.assume_yes:
            raise exceptions.CLIJSONFormatRequireAssumeYes()
        else:
            return f(cmd_args, *args, **kwargs)

    return new_f


def assert_attached(raise_custom_error_function=None):
    """Decorator asserting attached config.
    :param msg_function: Optional function to generate a custom message
    if raising an UnattachedError
    """

    def wrapper(f):
        @wraps(f)
        def new_f(args, cfg, **kwargs):
            if not _is_attached(cfg).is_attached:
                if raise_custom_error_function:
                    command = getattr(args, "command", "")
                    service_names = getattr(args, "service", "")
                    raise_custom_error_function(
                        command=command, service_names=service_names, cfg=cfg
                    )
                else:
                    raise exceptions.UnattachedError()
            return f(args, cfg=cfg, **kwargs)

        return new_f

    return wrapper


def assert_not_attached(f):
    """Decorator asserting unattached config."""

    @wraps(f)
    def new_f(args, cfg, **kwargs):
        if _is_attached(cfg).is_attached:
            machine_token_file = machine_token.get_machine_token_file()
            raise exceptions.AlreadyAttachedError(
                account_name=machine_token_file.account.get("name", "")
            )
        return f(args, cfg=cfg, **kwargs)

    return new_f


def assert_vulnerability_issue_valid(cmd):
    def wrapper(f):
        @wraps(f)
        def new_f(args, cfg, **kwargs):
            security_issue = getattr(args, "security_issue", "")
            if not re.match(CVE_OR_USN_REGEX, security_issue):
                raise exceptions.InvalidSecurityIssueIdFormat(
                    issue=security_issue,
                    cmd=cmd,
                )

            return f(args, cfg=cfg, **kwargs)

        return new_f

    return wrapper


def _raise_enable_disable_unattached_error(command, service_names, cfg):
    """Raises a custom error for enable/disable commands when unattached.

    Takes into consideration if the services exist or not, and notify the user
    accordingly."""
    (
        entitlements_found,
        entitlements_not_found,
    ) = entitlements.get_valid_entitlement_names(names=service_names, cfg=cfg)
    if entitlements_found and entitlements_not_found:
        raise exceptions.UnattachedMixedServicesError(
            valid_service=", ".join(entitlements_found),
            operation=command,
            invalid_service=", ".join(entitlements_not_found),
            service_msg="",
        )
    elif entitlements_found:
        raise exceptions.UnattachedValidServicesError(
            valid_service=", ".join(entitlements_found), operation=command
        )
    else:
        raise exceptions.UnattachedInvalidServicesError(
            operation=command,
            invalid_service=", ".join(entitlements_not_found),
            service_msg="",
        )


def post_cli_attach(cfg: UAConfig) -> None:
    machine_token_file = machine_token.get_machine_token_file(cfg)
    contract_name = machine_token_file.contract_name

    if contract_name:
        event.info(
            messages.ATTACH_SUCCESS_TMPL.format(contract_name=contract_name)
        )
    else:
        event.info(messages.ATTACH_SUCCESS_NO_CONTRACT_NAME)

    daemon.stop()
    daemon.cleanup(cfg)

    status_dict, _ret = actions.status(cfg)
    output = status.format_tabular(status_dict)
    event.info(util.handle_unicode_characters(output))
    event.process_events()


def configure_apt_proxy(
    cfg: UAConfig,
    scope: AptProxyScope,
    set_key: str,
    set_value: Optional[str],
) -> None:
    """
    Handles setting part the apt proxies - global and uaclient scoped proxies
    """
    if scope == AptProxyScope.GLOBAL:
        http_proxy = cfg.global_apt_http_proxy
        https_proxy = cfg.global_apt_https_proxy
    elif scope == AptProxyScope.UACLIENT:
        http_proxy = cfg.ua_apt_http_proxy
        https_proxy = cfg.ua_apt_https_proxy
    if "https" in set_key:
        https_proxy = set_value
    else:
        http_proxy = set_value
    setup_apt_proxy(
        http_proxy=http_proxy, https_proxy=https_proxy, proxy_scope=scope
    )


def run_spinner(stop_spinner, msg):
    spinner = itertools.cycle(["|", "/", "-", "\\"])
    while not stop_spinner.is_set():
        sys.stdout.write(
            "\r{}{}{}".format(msg, next(spinner), CLEAR_LINE_ANSI_CODE)
        )
        sys.stdout.flush()
        time.sleep(0.1)
        sys.stdout.write("\b")


def with_spinner(msg: Optional[str] = ""):
    def wrapper(f):
        @wraps(f)
        def new_f(args, *, cfg, **kwargs):
            if not sys.stdout.isatty():
                return f(args, cfg=cfg, **kwargs)

            stop_spinner = threading.Event()

            spinner_thread = threading.Thread(
                target=run_spinner, args=(stop_spinner, msg)
            )
            spinner_thread.start()

            retval = f(args, cfg=cfg, **kwargs)

            stop_spinner.set()
            spinner_thread.join()

            sys.stdout.write("\r" + CLEAR_LINE_ANSI_CODE)
            sys.stdout.flush()

            return retval

        return new_f

    return wrapper


def colorize_priority(priority):
    if priority == "low":
        return messages.TxtColor.INFOBLUE + priority + messages.TxtColor.ENDC
    elif priority == "medium":
        return (
            messages.TxtColor.WARNINGYELLOW + priority + messages.TxtColor.ENDC
        )
    elif priority == "high":
        return messages.TxtColor.ORANGE + priority + messages.TxtColor.ENDC
    elif priority == "critical":
        return messages.TxtColor.FAIL + priority + messages.TxtColor.ENDC
    else:
        return priority



import datetime
import json
import logging
import os
import re
import sys
import textwrap
import time
from functools import wraps
from typing import Any, Dict, List, Optional, Union  # noqa: F401

from uaclient import exceptions, messages
from uaclient.defaults import CONFIG_FIELD_ENVVAR_ALLOWLIST
from uaclient.types import MessagingOperations

DROPPED_KEY = object()


def replace_top_level_logger_name(name: str) -> str:
    """Replace the name of the root logger from __name__"""
    if name == "":
        return ""
    names = name.split(".")
    names[0] = "ubuntupro"
    return ".".join(names)


LOG = logging.getLogger(replace_top_level_logger_name(__name__))


class DatetimeAwareJSONEncoder(json.JSONEncoder):
    """A json.JSONEncoder subclass that writes out isoformat'd datetimes."""

    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        return super().default(o)


class DatetimeAwareJSONDecoder(json.JSONDecoder):
    """
    A JSONDecoder that parses some ISO datetime strings to datetime objects.

    Important note: the "some" is because we seem to only be able extend
    Python's json library in a way that lets us convert string values within
    JSON objects (e.g. '{"lastModified": "2019-07-25T14:35:51"}'). Strings
    outside of JSON objects (e.g. '"2019-07-25T14:35:51"') will not be passed
    through our decoder.

    (N.B. This will override any object_hook specified using arguments to it,
    or used in load or loads calls that specify this as the cls.)
    """

    def __init__(self, *args, **kwargs):
        if "object_hook" in kwargs:
            kwargs.pop("object_hook")
        super().__init__(*args, object_hook=self.object_hook, **kwargs)

    @staticmethod
    def object_hook(o):
        for key, value in o.items():
            if isinstance(value, str):
                try:
                    new_value = parse_rfc3339_date(
                        value
                    )  # type: Union[str, datetime.datetime]
                except ValueError:
                    # This isn't a string containing a valid ISO 8601 datetime
                    new_value = value
                o[key] = new_value
        return o


def retry(exception, retry_sleeps):
    """Decorator to retry on exception for retry_sleeps.

    @param retry_sleeps: List of sleep lengths to apply between
       retries. Specifying a list of [0.5, 1] tells subp to retry twice
       on failure; sleeping half a second before the first retry and 1 second
       before the second retry.
    @param exception: The exception class to catch and retry for the provided
       retry_sleeps. Any other exception types will not be caught by the
       decorator.
    """

    def wrapper(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            sleeps = retry_sleeps.copy()
            while True:
                try:
                    return f(*args, **kwargs)
                except exception as e:
                    if not sleeps:
                        raise e
                    LOG.debug(
                        "%s: Retrying %d more times.", str(e), len(sleeps)
                    )
                    time.sleep(sleeps.pop(0))

        return decorator

    return wrapper


def get_dict_deltas(
    orig_dict: Dict[str, Any], new_dict: Dict[str, Any], path: str = ""
) -> Dict[str, Any]:
    """Return a dictionary of delta between orig_dict and new_dict."""
    deltas = {}  # type: Dict[str, Any]
    for key, value in orig_dict.items():
        new_value = new_dict.get(key, DROPPED_KEY)
        key_path = key if not path else path + "." + key
        if isinstance(value, dict):
            if key in new_dict:
                sub_delta = get_dict_deltas(
                    value, new_dict[key], path=key_path
                )
                if sub_delta:
                    deltas[key] = sub_delta
            else:
                deltas[key] = DROPPED_KEY
        elif value != new_value:
            LOG.debug(
                "Contract value for '%s' changed to '%s'",
                key_path,
                str(new_value),
            )
            deltas[key] = new_value
    for key, value in new_dict.items():
        if key not in orig_dict:
            deltas[key] = value
    return deltas


def prompt_choices(msg: str = "", valid_choices: List[str] = []) -> str:
    """Interactive prompt message, returning a valid choice from msg.

    Expects a structured msg which designates choices with square brackets []
    around the characters which indicate a valid choice.

    Uppercase and lowercase responses are allowed. Loop on invalid choices.

    :return: Valid response character chosen.
    """
    from uaclient import event_logger

    event = event_logger.get_event_logger()
    value = ""
    error_msg = "{} is not one of: {}".format(
        value, ", ".join([choice.upper() for choice in valid_choices])
    )
    while True:
        event.info(msg)
        value = input("> ").lower()
        if value in valid_choices:
            break
        event.info(error_msg)
    return value


def prompt_for_confirmation(
    msg: str = "", assume_yes: bool = False, default: bool = False
) -> bool:
    """
    Display a confirmation prompt, returning a bool indicating the response

    :param msg: String custom prompt text to emit from input call.
    :param assume_yes: Boolean set True to skip confirmation input and return
        True.
    :param default: Boolean to return when user doesn't enter any text

    This function will only prompt a single time, and defaults to "no" (i.e. it
    returns False).
    """
    if assume_yes:
        return True
    if not msg:
        msg = messages.PROMPT_YES_NO
    value = input(msg).lower().strip()
    if value == "":
        return default
    if value in ["y", "yes"]:
        return True
    return False


def is_config_value_true(config: Dict[str, Any], path_to_value: str) -> bool:
    """Check if value parameter can be translated into a boolean 'True' value.

    @param config: A config dict representing
                   /etc/ubuntu-advantange/uaclient.conf
    @param path_to_value: The path from where the value parameter was
                          extracted.
    @return: A boolean value indicating if the value paramater corresponds
             to a 'True' boolean value.
    @raises exceptions.UbuntuProError when the value provide by the
            path_to_value parameter can not be translated into either
            a 'False' or 'True' boolean value.
    """
    value = config
    default_value = {}  # type: Any
    paths = path_to_value.split(".")
    leaf_value = paths[-1]
    for key in paths:
        if key == leaf_value:
            default_value = "false"

        if isinstance(value, dict):
            value = value.get(key, default_value)
        else:
            return False

    value_str = str(value)
    if value_str.lower() == "true":
        return True
    elif value_str.lower() == "false":
        return False
    else:
        raise exceptions.InvalidBooleanConfigValue(
            path_to_value=path_to_value,
            expected_value="boolean string: true or false",
            value=value_str,
        )


REDACT_SENSITIVE_LOGS = [
    r"(Bearer )[^\']+",
    r"(\'attach\', \')[^\']+",
    r"(\'machineToken\': \')[^\']+",
    r"(\'token\': \')[^\']+",
    r"(\'X-aws-ec2-metadata-token\': \')[^\']+",
    r"(.*\[PUT\] response.*api/token,.*data: ).*",
    r"(https://bearer:)[^\@]+",
    r"(/snap/bin/canonical-livepatch\s+enable\s+)[^\s]+",
    r"(Contract\s+value\s+for\s+'resourceToken'\s+changed\s+to\s+).*",
    r"(\'resourceToken\': \')[^\']+",
    r"(\'contractToken\': \')[^\']+",
    r"(https://contracts.canonical.com/v1/resources/livepatch\?token=)[^\s]+",
    r"(\"identityToken\": \")[^\"]+",
    r"(response:\s+http://metadata/computeMetadata/v1/instance/"
    "service-accounts.*data: ).*",
    r"(\'token\': \')[^\']+",
    r"(\'userCode\': \')[^\']+",
    r"(\'magic_token=)[^\']+",
    r"(--registration-key=\")[^\"]+",
    r"(--registration-key=\')[^\']+",
    r"(--registration-key=)[^ ]+",
    r"(--registration-key \")[^\"]+",
    r"(--registration-key \')[^\']+",
    r"(--registration-key )[^\s]+",
    r"(-p \")[^\"]+",
    r"(-p \')[^\']+",
    r"(-p )[^\s]+",
]


def redact_sensitive_logs(
    log, redact_regexs: List[str] = REDACT_SENSITIVE_LOGS
) -> str:
    """Redact known sensitive information from log content."""
    redacted_log = log
    for redact_regex in redact_regexs:
        redacted_log = re.sub(redact_regex, r"\g<1><REDACTED>", redacted_log)
    return redacted_log


def handle_message_operations(
    msg_ops: Optional[MessagingOperations], assume_yes: bool
) -> bool:
    """Emit messages to the console for user interaction

    :param msg_op: A list of strings or tuples. Any string items are printed.
        Any tuples will contain a callable and a dict of args to pass to the
        callable. Callables are expected to return True on success and
        False upon failure.

    :return: True upon success, False on failure.
    """
    if not msg_ops:
        return True

    for msg_op in msg_ops:
        if isinstance(msg_op, str):
            print(msg_op)
        else:  # Then we are a callable and dict of args
            functor, args = msg_op
            args["assume_yes"] = assume_yes
            if not functor(**args):
                return False
    return True


def parse_rfc3339_date(dt_str: str) -> datetime.datetime:
    """
    Parse a datestring in rfc3339 format. Originally written for compatibility
    with golang's time.MarshalJSON function. Also handles output of pythons
    isoformat datetime method.

    This drops subseconds.

    :param dt_str: a date string in rfc3339 format

    :return: datetime.datetime object of time represented by dt_str
    """
    # remove sub-seconds
    # Examples:
    #   Before: "2001-02-03T04:05:06.123456"
    #   After: "2001-02-03T04:05:06"
    #   Before: "2001-02-03T04:05:06.123456Z"
    #   After: "2001-02-03T04:05:06Z"
    #   Before: "2001-02-03T04:05:06.123456+09:00"
    #   After: "2001-02-03T04:05:06+09:00"
    dt_str_without_subseconds = re.sub(
        r"(\d{2}:\d{2}:\d{2})\.\d+", r"\g<1>", dt_str
    )
    # if there is no timezone info, assume UTC
    # Examples:
    #   Before: "2001-02-03T04:05:06"
    #   After: "2001-02-03T04:05:06Z"
    #   Before: "2001-02-03T04:05:06Z"
    #   After: "2001-02-03T04:05:06Z"
    #   Before: "2001-02-03T04:05:06+09:00"
    #   After: "2001-02-03T04:05:06+09:00"
    dt_str_with_z = re.sub(
        r"(\d{2}:\d{2}:\d{2})$", r"\g<1>Z", dt_str_without_subseconds
    )
    # replace Z with offset for UTC
    # Examples:
    #   Before: "2001-02-03T04:05:06Z"
    #   After: "2001-02-03T04:05:06+00:00"
    #   Before: "2001-02-03T04:05:06+09:00"
    #   After: "2001-02-03T04:05:06+09:00"
    dt_str_without_z = dt_str_with_z.replace("Z", "+00:00")
    # change offset format to not include colon `:`
    # Examples:
    #   Before: "2001-02-03T04:05:06+00:00"
    #   After: "2001-02-03T04:05:06+0000"
    #   Before: "2001-02-03T04:05:06+09:00"
    #   After: "2001-02-03T04:05:06+0900"
    dt_str_with_pythonish_tz = re.sub(
        r"(-|\+)(\d{2}):(\d{2})$", r"\g<1>\g<2>\g<3>", dt_str_without_z
    )
    return datetime.datetime.strptime(
        dt_str_with_pythonish_tz, "%Y-%m-%dT%H:%M:%S%z"
    )


def handle_unicode_characters(message: str) -> str:
    """
    Verify if the system can output unicode characters and if not,
    remove those characters from the message string.
    """
    if (
        sys.stdout.encoding is None
        or "UTF-8" not in sys.stdout.encoding.upper()
    ):
        # Replace our Unicode dash with an ASCII dash if we aren't going to be
        # writing to a utf-8 output; see
        # https://github.com/canonical/ubuntu-pro-client/issues/859
        message = message.replace("\u2014", "-")

        # Remove our unicode success/failure marks if we aren't going to be
        # writing to a utf-8 output; see
        # https://github.com/canonical/ubuntu-pro-client/issues/1463
        message = message.replace(messages.OKGREEN_CHECK + " ", "")
        message = message.replace(messages.FAIL_X + " ", "")

        # Now we remove any remaining unicode characters from the string
        message = message.encode("ascii", "ignore").decode()

    return message


def get_pro_environment():
    return {
        k: v
        for k, v in os.environ.items()
        if k.lower() in CONFIG_FIELD_ENVVAR_ALLOWLIST
        or k.startswith("UA_FEATURES")
        or k == "UA_CONFIG_FILE"
    }


def depth_first_merge_overlay_dict(base_dict, overlay_dict):
    """Merge the contents of overlay dict into base_dict not only on top-level
    keys, but on all on the depths of the overlay_dict object. For example,
    using these values as entries for the function:

    base_dict = {"a": 1, "b": {"c": 2, "d": 3}}
    overlay_dict = {"b": {"c": 10}}

    Should update base_dict into:

    {"a": 1, "b": {"c": 10, "d": 3}}

    @param base_dict: The dict to be updated
    @param overlay_dict: The dict with information to be added into base_dict
    """

    def update_dict_list(base_values, overlay_values, key):
        merge_id_key_map = {
            "availableResources": "name",
            "resourceEntitlements": "type",
            "overrides": "selector",
        }
        values_to_append = []
        id_key = merge_id_key_map.get(key)
        for overlay_value in overlay_values:
            was_replaced = False
            for base_value_idx, base_value in enumerate(base_values):
                if base_value.get(id_key) == overlay_value.get(id_key):
                    depth_first_merge_overlay_dict(base_value, overlay_value)
                    was_replaced = True

            if not was_replaced:
                values_to_append.append(overlay_value)

        base_values.extend(values_to_append)

    for key, value in overlay_dict.items():
        base_value = base_dict.get(key)
        if isinstance(base_value, dict) and isinstance(value, dict):
            depth_first_merge_overlay_dict(base_dict[key], value)
        elif isinstance(base_value, list) and isinstance(value, list):
            if len(base_value) and isinstance(base_value[0], dict):
                update_dict_list(base_dict[key], value, key=key)
            else:
                """
                Most other lists which aren't lists of dicts are lists of
                strs. Replace that list # with the overlay value."""
                base_dict[key] = value
        else:
            base_dict[key] = value


ARCH_ALIASES = {
    "x86_64": "amd64",
    "i686": "i386",
    "ppc64le": "ppc64el",
    "aarch64": "arm64",
    "armv7l": "armhf",
}


def standardize_arch_name(arch: str) -> str:
    arch_lower = arch.lower()
    return ARCH_ALIASES.get(arch_lower, arch_lower)


def deduplicate_arches(arches: List[str]) -> List[str]:
    deduplicated_arches = set()
    for arch in arches:
        deduplicated_arches.add(standardize_arch_name(arch))
    return sorted(list(deduplicated_arches))


def we_are_currently_root() -> bool:
    return os.getuid() == 0


def set_filename_extension(filename: str, new_extension: str) -> str:
    name, _extension = os.path.splitext(filename)
    return name + "." + new_extension


def create_package_list_str(
    package_list: List[str],
):
    return (
        "\n".join(
            textwrap.wrap(
                " ".join(package_list),
                width=80,
                break_long_words=False,
                break_on_hyphens=False,
                initial_indent="  ",
                subsequent_indent="  ",
            )
        )
        + "\n"
    )



import logging
from typing import Any, Dict, Optional, Tuple

from uaclient import (
    api,
    event_logger,
    exceptions,
    http,
    livepatch,
    messages,
    snap,
    system,
    util,
)
from uaclient.entitlements.base import EntitlementWithMessage, UAEntitlement
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.types import StaticAffordance

LIVEPATCH_RETRIES = [0.5, 1.0]

ERROR_MSG_MAP = {
    "Unknown Auth-Token": "Invalid Auth-Token provided to livepatch.",
    "unsupported kernel": "Your running kernel is not supported by Livepatch.",
}

event = event_logger.get_event_logger()
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class LivepatchEntitlement(UAEntitlement):
    help_doc_url = messages.urls.LIVEPATCH_HOME_PAGE
    name = "livepatch"
    title = messages.LIVEPATCH_TITLE
    description = messages.LIVEPATCH_DESCRIPTION
    help_text = messages.LIVEPATCH_HELP_TEXT
    affordance_check_kernel_min_version = False
    affordance_check_kernel_flavor = False
    # we do want to check series because livepatch errors on non-lts releases
    affordance_check_series = True
    # we still need to check arch because the livepatch-client is not built
    # for all arches
    affordance_check_arch = True

    @property
    def incompatible_services(self) -> Tuple[EntitlementWithMessage, ...]:
        from uaclient.entitlements.fips import FIPSEntitlement
        from uaclient.entitlements.realtime import RealtimeKernelEntitlement

        return (
            EntitlementWithMessage(
                FIPSEntitlement, messages.LIVEPATCH_INVALIDATES_FIPS
            ),
            EntitlementWithMessage(
                RealtimeKernelEntitlement,
                messages.REALTIME_LIVEPATCH_INCOMPATIBLE,
            ),
        )

    @property
    def static_affordances(self) -> Tuple[StaticAffordance, ...]:
        # Use a lambda so we can mock system.is_container in tests
        from uaclient.entitlements.fips import FIPSEntitlement

        fips_ent = FIPSEntitlement(cfg=self.cfg)

        is_fips_enabled = bool(
            fips_ent.application_status()[0] == ApplicationStatus.ENABLED
        )

        return (
            (
                messages.SERVICE_ERROR_INSTALL_ON_CONTAINER.format(
                    title=self.title
                ),
                lambda: system.is_container()
                or system.get_virt_type() == "wsl",
                False,
            ),
            (
                messages.LIVEPATCH_ERROR_WHEN_FIPS_ENABLED,
                lambda: is_fips_enabled,
                False,
            ),
        )

    def enable_steps(self) -> int:
        return 2

    def disable_steps(self) -> int:
        return 1

    def _perform_enable(self, progress: api.ProgressWrapper) -> bool:
        """Enable specific entitlement.

        @return: True on success, False otherwise.
        """
        progress.progress(messages.INSTALLING_LIVEPATCH)

        if not snap.is_snapd_installed():
            progress.emit(
                "info", messages.INSTALLING_PACKAGES.format(packages="snapd")
            )
            snap.install_snapd()

        if not snap.is_snapd_installed_as_a_snap():
            progress.emit(
                "info",
                messages.INSTALLING_PACKAGES.format(packages="snapd snap"),
            )
            try:
                snap.install_snap("snapd")
            except exceptions.ProcessExecutionError as e:
                LOG.warning("Failed to install snapd as a snap", exc_info=e)
                progress.emit(
                    "info",
                    messages.EXECUTING_COMMAND_FAILED.format(
                        command="snap install snapd"
                    ),
                )

        snap.run_snapd_wait_cmd(progress)

        try:
            snap.refresh_snap("snapd")
        except exceptions.ProcessExecutionError as e:
            LOG.warning("Failed to refresh snapd snap", exc_info=e)
            event.info(
                messages.EXECUTING_COMMAND_FAILED.format(
                    command="snap refresh snapd"
                )
            )

        http_proxy = http.validate_proxy(
            "http", self.cfg.http_proxy, http.PROXY_VALIDATION_SNAP_HTTP_URL
        )
        https_proxy = http.validate_proxy(
            "https", self.cfg.https_proxy, http.PROXY_VALIDATION_SNAP_HTTPS_URL
        )
        snap.configure_snap_proxy(
            http_proxy=http_proxy,
            https_proxy=https_proxy,
            retry_sleeps=snap.SNAP_INSTALL_RETRIES,
        )
        if not livepatch.is_livepatch_installed():
            progress.emit(
                "info",
                messages.INSTALLING_PACKAGES.format(
                    packages="canonical-livepatch snap"
                ),
            )
            try:
                snap.install_snap("canonical-livepatch")
            except exceptions.ProcessExecutionError as e:
                raise exceptions.ErrorInstallingLivepatch(error_msg=str(e))

        livepatch.configure_livepatch_proxy(http_proxy, https_proxy)

        return self.setup_livepatch_config(
            progress, process_directives=True, process_token=True
        )

    def setup_livepatch_config(
        self,
        progress: api.ProgressWrapper,
        process_directives: bool = True,
        process_token: bool = True,
    ) -> bool:
        """Processs configuration setup for livepatch directives.

        :param process_directives: Boolean set True when directives should be
            processsed.
        :param process_token: Boolean set True when token should be
            processsed.
        """
        progress.progress(messages.SETTING_UP_LIVEPATCH)

        entitlement_cfg = self.machine_token_file.entitlements().get(self.name)
        if process_directives:
            try:
                process_config_directives(entitlement_cfg)
            except exceptions.ProcessExecutionError as e:
                LOG.error(str(e), exc_info=e)
                progress.emit(
                    "info",
                    messages.LIVEPATCH_UNABLE_TO_CONFIGURE.format(
                        error_msg=str(e)
                    ),
                )
                return False
        if process_token:
            livepatch_token = entitlement_cfg.get("resourceToken")
            if not livepatch_token:
                LOG.debug(
                    "No specific resourceToken present. Using machine token as"
                    " %s credentials",
                    self.title,
                )
                livepatch_token = self.machine_token_file.machine_token[
                    "machineToken"
                ]
            application_status, _details = self.application_status()
            if application_status != ApplicationStatus.DISABLED:
                LOG.info("Disabling livepatch before re-enabling")
                progress.emit("info", messages.LIVEPATCH_DISABLE_REATTACH)
                try:
                    system.subp([livepatch.LIVEPATCH_CMD, "disable"])
                except exceptions.ProcessExecutionError as e:
                    LOG.error(str(e), exc_info=e)
                    return False
            try:
                system.subp(
                    [livepatch.LIVEPATCH_CMD, "enable", livepatch_token],
                    capture=True,
                )
            except exceptions.ProcessExecutionError as e:
                msg = messages.LIVEPATCH_UNABLE_TO_ENABLE
                for error_message, print_message in ERROR_MSG_MAP.items():
                    if error_message in str(e):
                        msg += print_message
                        break
                if msg == messages.LIVEPATCH_UNABLE_TO_ENABLE:
                    msg += str(e)
                progress.emit("info", msg)
                return False
        return True

    def _perform_disable(self, progress: api.ProgressWrapper):
        """Disable specific entitlement

        @return: True on success, False otherwise.
        """
        if not livepatch.is_livepatch_installed():
            return True
        cmd = [livepatch.LIVEPATCH_CMD, "disable"]
        progress.progress(
            messages.EXECUTING_COMMAND.format(command=" ".join(cmd))
        )
        system.subp(cmd, capture=True)
        return True

    def application_status(
        self,
    ) -> Tuple[ApplicationStatus, Optional[messages.NamedMessage]]:
        status = (ApplicationStatus.ENABLED, None)

        if not livepatch.is_livepatch_installed():
            return (ApplicationStatus.DISABLED, messages.LIVEPATCH_NOT_ENABLED)

        try:
            livepatch_status = livepatch.status()
        except exceptions.ProcessExecutionError as e:
            return (
                ApplicationStatus.WARNING,
                messages.LIVEPATCH_CLIENT_FAILURE_WARNING.format(
                    livepatch_error=e.stderr,
                ),
            )

        if livepatch_status is None:
            # TODO(May want to parse INACTIVE/failure assessment)
            return (
                ApplicationStatus.DISABLED,
                messages.LIVEPATCH_APPLICATION_STATUS_CLIENT_FAILURE,
            )
        return status

    def enabled_warning_status(
        self,
    ) -> Tuple[bool, Optional[messages.NamedMessage]]:
        support = livepatch.on_supported_kernel()
        if support == livepatch.LivepatchSupport.UNSUPPORTED:
            kernel_info = system.get_kernel_info()
            return (
                True,
                messages.LIVEPATCH_KERNEL_NOT_SUPPORTED.format(
                    version=kernel_info.uname_release,
                    arch=kernel_info.uname_machine_arch,
                ),
            )
        if support == livepatch.LivepatchSupport.KERNEL_EOL:
            kernel_info = system.get_kernel_info()
            return (
                True,
                messages.LIVEPATCH_KERNEL_EOL.format(
                    version=kernel_info.uname_release,
                    arch=kernel_info.uname_machine_arch,
                ),
            )
        if support == livepatch.LivepatchSupport.KERNEL_UPGRADE_REQUIRED:
            return (
                True,
                messages.LIVEPATCH_KERNEL_UPGRADE_REQUIRED,
            )

        # if on_supported_kernel returns UNKNOWN we default to no warning
        # because there would be no way for a user to resolve the warning
        return False, None

    def status_description_override(self):
        if (
            livepatch.on_supported_kernel()
            == livepatch.LivepatchSupport.UNSUPPORTED
            and not system.is_container()
        ):
            return messages.LIVEPATCH_KERNEL_NOT_SUPPORTED_DESCRIPTION
        return None

    def process_contract_deltas(
        self,
        orig_access: Dict[str, Any],
        deltas: Dict[str, Any],
        allow_enable: bool = False,
        verbose: bool = True,
    ) -> bool:
        """Process any contract access deltas for this entitlement.

        :param orig_access: Dictionary containing the original
            resourceEntitlement access details.
        :param deltas: Dictionary which contains only the changed access keys
        and values.
        :param allow_enable: Boolean set True if allowed to perform the enable
            operation. When False, a message will be logged to inform the user
            about the recommended enabled service.
        :param verbose: If True, display output to stdout

        :return: True when delta operations are processed; False when noop.
        """
        if super().process_contract_deltas(
            orig_access, deltas, allow_enable, verbose=verbose
        ):
            return True  # Already processed parent class deltas

        delta_entitlement = deltas.get("entitlement", {})
        process_enable_default = delta_entitlement.get("obligations", {}).get(
            "enableByDefault", False
        )

        if process_enable_default:
            enable_success, _ = self.enable(api.ProgressWrapper())
            return enable_success

        application_status, _ = self.application_status()
        if application_status == ApplicationStatus.DISABLED:
            return False  # only operate on changed directives when ACTIVE
        delta_directives = delta_entitlement.get("directives", {})
        supported_deltas = set(["caCerts", "remoteServer"])
        process_directives = bool(
            supported_deltas.intersection(delta_directives)
        )
        process_token = bool(deltas.get("resourceToken", False))
        if any([process_directives, process_token]):
            LOG.info(
                "New livepatch directives or token. running "
                "setup_livepatch_config"
            )

            if verbose:
                event.info(
                    messages.SERVICE_UPDATING_CHANGED_DIRECTIVES.format(
                        service=self.name
                    )
                )
            return self.setup_livepatch_config(
                progress=api.ProgressWrapper(),
                process_directives=process_directives,
                process_token=process_token,
            )
        return True


def process_config_directives(cfg):
    """Process livepatch configuration directives.

    We process caCerts before remoteServer because changing remote-server
    in the canonical-livepatch CLI performs a PUT against the new server name.
    If new caCerts were required for the new remoteServer, this
    canonical-livepatch client PUT could fail on unmatched old caCerts.

    @raises: ProcessExecutionError if unable to configure livepatch.
    """
    if not cfg:
        return
    directives = cfg.get("entitlement", {}).get("directives", {})
    ca_certs = directives.get("caCerts")
    if ca_certs:
        system.subp(
            [
                livepatch.LIVEPATCH_CMD,
                "config",
                "ca-certs={}".format(ca_certs),
            ],
            capture=True,
        )
    remote_server = directives.get("remoteServer", "")
    if remote_server.endswith("/"):
        remote_server = remote_server[:-1]
    if remote_server:
        system.subp(
            [
                livepatch.LIVEPATCH_CMD,
                "config",
                "remote-server={}".format(remote_server),
            ],
            capture=True,
        )



"""
Project-wide default settings

These are in their own file so they can be imported by setup.py before we have
any of our dependencies installed.
"""

import os

# Base directories
UAC_ETC_PATH = "/etc/ubuntu-advantage"
UAC_RUN_PATH = "/run/ubuntu-advantage"
DEFAULT_DATA_DIR = "/var/lib/ubuntu-advantage"
DEFAULT_LOG_DIR = "/var/log"


# Relative paths
MACHINE_TOKEN_FILE = "machine-token.json"
CONFIG_FILE = "uaclient.conf"
USER_CONFIG_FILE = "user-config.json"
CANDIDATE_VERSION_FILE = "candidate-version"
DEFAULT_LOG_FILE_BASE_NAME = "ubuntu-advantage"
PRIVATE_SUBDIR = "private"
MESSAGES_SUBDIR = "messages"
INTERFACE_FILES_SUBDIR = "interfaces"
USER_CACHE_SUBDIR = "ubuntu-pro"
NOTICES_SUBDIR = "notices"
PRIVATE_ESM_CACHE_SUBDIR = "apt-esm"
VULNERABILITY_SUBDIR = "vulnerability-data"
VULNERABILITY_DATA_CACHE = "vulnerability-cache.json"
VULNERABILITY_RESULT_CACHE = "vulnerability-result.json"
VULNERABILITY_ETAG_CACHE = "vulnerability-etag"
VULNERABILITY_DPKG_STATUS_DATE_CACHE = "vulnerability-dpkg-status-date"

DEFAULT_PRIVATE_MACHINE_TOKEN_PATH = os.path.join(
    DEFAULT_DATA_DIR, PRIVATE_SUBDIR, MACHINE_TOKEN_FILE
)
DEFAULT_PRIVATE_DATA_DIR = os.path.join(DEFAULT_DATA_DIR, PRIVATE_SUBDIR)
MESSAGES_DIR = os.path.join(DEFAULT_DATA_DIR, MESSAGES_SUBDIR)
INTERFACE_FILES_DIR = os.path.join(DEFAULT_DATA_DIR, INTERFACE_FILES_SUBDIR)
DEFAULT_CONFIG_FILE = os.path.join(UAC_ETC_PATH, CONFIG_FILE)
CANDIDATE_CACHE_PATH = os.path.join(UAC_RUN_PATH, CANDIDATE_VERSION_FILE)
DEFAULT_USER_CONFIG_JSON_FILE = os.path.join(
    DEFAULT_DATA_DIR, USER_CONFIG_FILE
)
DEFAULT_LOG_PREFIX = os.path.join(DEFAULT_LOG_DIR, DEFAULT_LOG_FILE_BASE_NAME)
ESM_APT_ROOTDIR = os.path.join(DEFAULT_DATA_DIR, PRIVATE_ESM_CACHE_SUBDIR)
NOTICES_PERMANENT_DIRECTORY = os.path.join(DEFAULT_DATA_DIR, NOTICES_SUBDIR)
NOTICES_TEMPORARY_DIRECTORY = os.path.join(UAC_RUN_PATH, NOTICES_SUBDIR)
APT_NEWS_PATH = os.path.join(UAC_RUN_PATH, "apt-news")

# URLs
BASE_CONTRACT_URL = "https://contracts.canonical.com"
BASE_SECURITY_URL = "https://ubuntu.com/security"
BASE_LIVEPATCH_URL = "https://livepatch.canonical.com"
BASE_VULNERABILITY_DATA_URL = "https://security-metadata.canonical.com/oval/"
APT_NEWS_URL = "https://motd.ubuntu.com/aptnews.json"

# VULNERABILITY DATA
VULNERABILITY_DATA_TMPL = "com.ubuntu.{series}.pkg.json.xz"
VULNERABILITY_CACHE_PATH = os.path.join(DEFAULT_DATA_DIR, VULNERABILITY_SUBDIR)

PRINT_WRAP_WIDTH = 80
CONTRACT_EXPIRY_GRACE_PERIOD_DAYS = 14
CONTRACT_EXPIRY_PENDING_DAYS = 20
ATTACH_FAIL_DATE_FORMAT = "%B %d, %Y"

DEFAULT_LOG_FORMAT = (
    "%(asctime)s - %(filename)s:(%(lineno)d) [%(levelname)s]: %(message)s"
)

CONFIG_DEFAULTS = {
    "contract_url": BASE_CONTRACT_URL,
    "security_url": BASE_SECURITY_URL,
    "data_dir": DEFAULT_DATA_DIR,
    "log_level": "debug",
    "log_file": "{}.log".format(DEFAULT_LOG_PREFIX),
}

CONFIG_FIELD_ENVVAR_ALLOWLIST = [
    "ua_data_dir",
    "ua_log_file",
    "ua_log_level",
    "ua_security_url",
]

ROOT_READABLE_MODE = 0o600
WORLD_READABLE_MODE = 0o644

CLOUD_BUILD_INFO = "/etc/cloud/build.info"
SSL_CERTS_PATH = "/etc/ssl/certs/ca-certificates.crt"

# used by apport, collect-logs, and tests
APPARMOR_PROFILES = [
    "/etc/apparmor.d/ubuntu_pro_apt_news",
    "/etc/apparmor.d/ubuntu_pro_esm_cache",
]



import mock
import pytest

from uaclient import exceptions, gpg
from uaclient.testing import helpers


class TestExportGPGKey:
    @pytest.mark.parametrize(
        ["source_exists", "expected_raises"],
        [
            (True, helpers.does_not_raise()),
            (False, pytest.raises(exceptions.GPGKeyNotFound)),
        ],
    )
    @mock.patch("uaclient.gpg.shutil.copy")
    @mock.patch("uaclient.gpg.os.chmod")
    @mock.patch("uaclient.gpg.os.path.exists")
    def test_export_gpg_key(
        self,
        m_exists,
        m_chmod,
        m_copy,
        source_exists,
        expected_raises,
    ):
        m_exists.return_value = source_exists
        with expected_raises:
            gpg.export_gpg_key(mock.sentinel.source, mock.sentinel.dest)
            assert m_copy.call_args_list == [
                mock.call(mock.sentinel.source, mock.sentinel.dest)
            ]
            assert m_chmod.call_args_list == [
                mock.call(mock.sentinel.dest, 0o644)
            ]



from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, IntDataValue
from uaclient.security_status import get_installed_packages_by_origin


class PackageSummary(DataObject):
    fields = [
        Field(
            "num_installed_packages",
            IntDataValue,
            doc="Total count of installed packages",
        ),
        Field(
            "num_esm_apps_packages",
            IntDataValue,
            doc="Count of packages installed from ``esm-apps``",
        ),
        Field(
            "num_esm_infra_packages",
            IntDataValue,
            doc="Count of packages installed from ``esm-infra``",
        ),
        Field(
            "num_main_packages",
            IntDataValue,
            doc="Count of packages installed from ``main``",
        ),
        Field(
            "num_multiverse_packages",
            IntDataValue,
            doc="Count of packages installed from ``multiverse``",
        ),
        Field(
            "num_restricted_packages",
            IntDataValue,
            doc="Count of packages installed from ``restricted``",
        ),
        Field(
            "num_third_party_packages",
            IntDataValue,
            doc="Count of packages installed from third party sources",
        ),
        Field(
            "num_universe_packages",
            IntDataValue,
            doc="Count of packages installed from ``universe``",
        ),
        Field(
            "num_unknown_packages",
            IntDataValue,
            doc="Count of packages installed from unknown sources",
        ),
    ]

    def __init__(
        self,
        num_installed_packages: int,
        num_esm_apps_packages: int,
        num_esm_infra_packages: int,
        num_main_packages: int,
        num_multiverse_packages: int,
        num_restricted_packages: int,
        num_third_party_packages: int,
        num_universe_packages: int,
        num_unknown_packages: int,
    ):
        self.num_installed_packages = num_installed_packages
        self.num_esm_apps_packages = num_esm_apps_packages
        self.num_esm_infra_packages = num_esm_infra_packages
        self.num_main_packages = num_main_packages
        self.num_multiverse_packages = num_multiverse_packages
        self.num_restricted_packages = num_restricted_packages
        self.num_third_party_packages = num_third_party_packages
        self.num_universe_packages = num_universe_packages
        self.num_unknown_packages = num_unknown_packages


class PackageSummaryResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "summary",
            PackageSummary,
            doc=("Summary of all installed packages"),
        )
    ]

    def __init__(self, summary):
        self.summary = summary


def summary() -> PackageSummaryResult:
    return _summary(UAConfig())


def _summary(cfg: UAConfig) -> PackageSummaryResult:
    """
    This endpoint shows a summary of installed packages in the system,
    categorized by origin.
    """
    packages = get_installed_packages_by_origin()
    summary = PackageSummary(
        num_installed_packages=len(packages["all"]),
        num_esm_apps_packages=len(packages["esm-apps"]),
        num_esm_infra_packages=len(packages["esm-infra"]),
        num_main_packages=len(packages["main"]),
        num_multiverse_packages=len(packages["multiverse"]),
        num_restricted_packages=len(packages["restricted"]),
        num_third_party_packages=len(packages["third-party"]),
        num_universe_packages=len(packages["universe"]),
        num_unknown_packages=len(packages["unknown"]),
    )
    return PackageSummaryResult(summary=summary)


endpoint = APIEndpoint(
    version="v1",
    name="PackageSummary",
    fn=_summary,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.12",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.packages.summary.v1 import summary

result = summary()
""",  # noqa: E501
    "result_class": PackageSummaryResult,
    "exceptions": [],
    "example_cli": "pro api u.pro.packages.summary.v1",
    "example_json": """
{
    "summary":{
        "num_installed_packages": 1,
        "num_esm_apps_packages": 2,
        "num_esm_infra_packages": 3,
        "num_main_packages": 4,
        "num_multiverse_packages": 5,
        "num_restricted_packages": 6,
        "num_third_party_packages": 7,
        "num_universe_packages": 8,
        "num_unknown_packages": 9,
    },
}
""",
}



from typing import List, Optional

from uaclient import apt, messages, util
from uaclient.api.u.pro.security.fix._common import FixStatus, status_message
from uaclient.api.u.pro.security.fix._common.plan.v1 import (
    FixPlanAptUpgradeStep,
    FixPlanAttachStep,
    FixPlanEnableStep,
    FixPlanNoOpStatus,
    FixPlanNoOpStep,
    FixPlanResult,
    FixPlanWarningPackageCannotBeInstalled,
    FixPlanWarningSecurityIssueNotFixed,
)
from uaclient.data_types import DataObject, Field, StringDataValue, data_list


class UpgradedPackage(DataObject):
    fields = [
        Field("name", StringDataValue, doc="The name of the package"),
        Field(
            "version",
            StringDataValue,
            doc="The version that the package was upgraded to",
        ),
        Field(
            "pocket",
            StringDataValue,
            doc="The pocket which contained the package upgrade",
        ),
    ]

    def __init__(self, name: str, version: str, pocket: str):
        self.name = name
        self.version = version
        self.pocket = pocket


class FailedUpgrade(DataObject):
    fields = [
        Field("name", StringDataValue, doc="The name of the package"),
        Field(
            "pocket",
            StringDataValue,
            required=False,
            doc="The pocket which contained the package upgrade",
        ),
    ]

    def __init__(self, name: str, pocket: Optional[str] = None):
        self.name = name
        self.pocket = pocket


class FixExecuteError(DataObject):
    fields = [
        Field("error_type", StringDataValue, doc="The type of the error"),
        Field(
            "reason", StringDataValue, doc="The reason why the error occurred"
        ),
        Field(
            "failed_upgrades",
            data_list(FailedUpgrade),
            required=False,
            doc="A list of ``FailedUpgrade`` objects",
        ),
    ]

    def __init__(
        self,
        error_type: str,
        reason: str,
        failed_upgrades: Optional[List[FailedUpgrade]] = None,
    ):
        self.error_type = error_type
        self.reason = reason
        self.failed_upgrades = failed_upgrades


class FixExecuteResult(DataObject):
    fields = [
        Field("title", StringDataValue, doc="The title of the CVE"),
        Field(
            "description",
            StringDataValue,
            required=False,
            doc="The description of the CVE",
        ),
        Field("status", StringDataValue, doc="The status of fixing the CVE"),
        Field(
            "upgraded_packages",
            data_list(UpgradedPackage),
            required=False,
            doc="A list of ``UpgradedPackage`` objects",
        ),
        Field(
            "errors",
            data_list(FixExecuteError),
            required=False,
            doc="A list of ``FixExecuteError`` objects",
        ),
    ]

    def __init__(
        self,
        title: str,
        status: str,
        description: Optional[str] = None,
        upgraded_packages: Optional[List[UpgradedPackage]] = None,
        errors: Optional[List[FixExecuteError]] = None,
    ):
        self.title = title
        self.description = description
        self.status = status
        self.upgraded_packages = upgraded_packages
        self.errors = errors


class ExecuteContext:
    def __init__(self):
        self.require_enable = False
        self.require_attach = False
        self.status = FixStatus.SYSTEM_NON_VULNERABLE.value.msg
        self.upgraded_pkgs = []  # type: List[UpgradedPackage]
        self.errors = []  # type: List[FixExecuteError]


def _handle_error(
    execute_context: ExecuteContext, security_issue: FixPlanResult
):
    if security_issue.error:
        execute_context.errors.append(
            FixExecuteError(
                error_type=security_issue.error.code or "unexpected-error",
                reason=security_issue.error.msg,
            )
        )
        execute_context.status = "error"


def _handle_security_issue_not_fixed(
    execute_context: ExecuteContext,
    warning: FixPlanWarningSecurityIssueNotFixed,
):
    execute_context.errors.append(
        FixExecuteError(
            error_type=warning.warning_type,
            reason=status_message(warning.data.status),
            failed_upgrades=[
                FailedUpgrade(name=pkg) for pkg in warning.data.source_packages
            ],
        )
    )
    execute_context.status = FixStatus.SYSTEM_STILL_VULNERABLE.value.msg


def _handle_package_cannot_be_installed(
    execute_context: ExecuteContext,
    warning: FixPlanWarningPackageCannotBeInstalled,
):
    execute_context.errors.append(
        FixExecuteError(
            error_type=warning.warning_type,
            reason=messages.FIX_CANNOT_INSTALL_PACKAGE.format(
                package=warning.data.binary_package,
                version=warning.data.binary_package_version,
            ),
            failed_upgrades=[
                FailedUpgrade(
                    name=warning.data.binary_package,
                    pocket=warning.data.pocket,
                )
            ],
        )
    )
    execute_context.status = FixStatus.SYSTEM_STILL_VULNERABLE.value.msg


def _handle_attach(execute_context: ExecuteContext, step: FixPlanAttachStep):
    execute_context.errors.append(
        FixExecuteError(
            error_type="fix-requires-attach",
            reason=messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION,
            failed_upgrades=[
                FailedUpgrade(name=pkg, pocket=step.data.required_service)
                for pkg in step.data.source_packages
            ],
        )
    )
    execute_context.require_attach = True
    execute_context.status = FixStatus.SYSTEM_STILL_VULNERABLE.value.msg


def _handle_enable(execute_context: ExecuteContext, step: FixPlanEnableStep):
    if execute_context.require_attach:
        return

    execute_context.errors.append(
        FixExecuteError(
            error_type="fix-requires-enable",
            reason=messages.SECURITY_SERVICE_DISABLED.format(
                service=step.data.service
            ),
            failed_upgrades=[
                FailedUpgrade(name=pkg, pocket=step.data.service)
                for pkg in step.data.source_packages
            ],
        )
    )
    execute_context.require_enable = True
    execute_context.status = FixStatus.SYSTEM_STILL_VULNERABLE.value.msg


def _handle_apt_upgrade(
    execute_context: ExecuteContext, step: FixPlanAptUpgradeStep
):
    if execute_context.require_attach or execute_context.require_enable:
        return

    if not step.data.binary_packages:
        return

    if not util.we_are_currently_root():
        execute_context.errors.append(
            FixExecuteError(
                error_type="fix-require-root",
                reason=messages.SECURITY_APT_NON_ROOT,
                failed_upgrades=[
                    FailedUpgrade(name=pkg, pocket=step.data.pocket)
                    for pkg in step.data.source_packages
                ],
            )
        )
        execute_context.status = "error"
        return

    try:
        apt.run_apt_update_command()
        apt.run_apt_command(
            cmd=["apt-get", "install", "--only-upgrade", "-y"]
            + step.data.binary_packages,
            override_env_vars={"DEBIAN_FRONTEND": "noninteractive"},
        )

        for pkg in step.data.binary_packages:
            pkg_version = apt.get_pkg_version(pkg)

            if pkg_version:
                execute_context.upgraded_pkgs.append(
                    UpgradedPackage(
                        name=pkg,
                        version=pkg_version,
                        pocket=step.data.pocket,
                    )
                )

    except Exception as e:
        msg = getattr(e, "msg", str(e))
        execute_context.status = FixStatus.SYSTEM_STILL_VULNERABLE.value.msg
        execute_context.errors.append(
            FixExecuteError(
                error_type="fix-error-installing-pkg",
                reason=msg,
                failed_upgrades=[
                    FailedUpgrade(name=pkg, pocket=step.data.pocket)
                    for pkg in step.data.source_packages
                ],
            )
        )


def _handle_noop(execute_context: ExecuteContext, step: FixPlanNoOpStep):
    if step.data.status == FixPlanNoOpStatus.NOT_AFFECTED.value:
        execute_context.status = FixStatus.SYSTEM_NOT_AFFECTED.value.msg


def _execute_fix(security_issue: FixPlanResult) -> FixExecuteResult:
    execute_context = ExecuteContext()

    if security_issue.error:
        _handle_error(execute_context, security_issue)

    if security_issue.warnings:
        for warning in security_issue.warnings:
            if isinstance(warning, FixPlanWarningSecurityIssueNotFixed):
                _handle_security_issue_not_fixed(execute_context, warning)
            elif isinstance(warning, FixPlanWarningPackageCannotBeInstalled):
                _handle_package_cannot_be_installed(execute_context, warning)

    if security_issue.plan:
        for step in security_issue.plan:
            if isinstance(step, FixPlanAttachStep):
                _handle_attach(execute_context, step)
            elif isinstance(step, FixPlanEnableStep):
                _handle_enable(execute_context, step)
            elif isinstance(step, FixPlanAptUpgradeStep):
                _handle_apt_upgrade(execute_context, step)
            elif isinstance(step, FixPlanNoOpStep):
                _handle_noop(execute_context, step)

    return FixExecuteResult(
        title=security_issue.title,
        description=security_issue.description,
        status=execute_context.status,
        upgraded_packages=execute_context.upgraded_pkgs,
        errors=None if not execute_context.errors else execute_context.errors,
    )



import abc
import os
import re
import sys
import textwrap
from enum import Enum
from typing import Any, Dict, List, Optional  # noqa: F401

from uaclient.config import UAConfig
from uaclient.messages import TxtColor

COLOR_FORMATTING_PATTERN = r"\033\[.*?m"
LINK_START_PATTERN = r"\033]8;;.+?\033\\+"
LINK_END = "\033]8;;\033\\"
UTF8_ALTERNATIVES = {
    "—": "-",
    "✘": "x",
    "✔": "*",
}  # type: Dict[str, str]


class ContentAlignment(Enum):
    LEFT = "l"
    RIGHT = "r"


# Class attributes and methods so we don't need singletons or globals for this
class ProOutputFormatterConfig:
    use_utf8 = True
    use_color = True

    # Initializing the class after the import is useful for unit testing
    @classmethod
    def init(cls, cfg: UAConfig):
        cls.use_utf8 = (
            sys.stdout.encoding is not None
            and "UTF-8" in sys.stdout.encoding.upper()
        )

        cls.use_color = sys.stdout.isatty() and os.getenv("NO_COLOR") is None

    @classmethod
    def disable_color(cls) -> None:
        cls.use_color = False


ProOutputFormatterConfig.init(cfg=UAConfig())


def create_link(text: str, url: str) -> str:
    return "\033]8;;{url}\033\\{text}\033]8;;\033\\".format(url=url, text=text)


def real_len(text: str) -> int:
    # ignore colors if existing
    result = re.sub(COLOR_FORMATTING_PATTERN, "", text)
    # Ignore link control characters and metadata
    result = re.sub(LINK_START_PATTERN, "", result)
    result = result.replace(LINK_END, "")

    return len(result)


def _get_default_length():
    if sys.stdout.isatty():
        return os.get_terminal_size().columns
    # If you're not in a tty, we don't care about string length
    # If you have a thousand characters line, well, wow
    return 999


def process_formatter_config(text: str) -> str:
    output = text
    if not ProOutputFormatterConfig.use_color:
        output = re.sub(COLOR_FORMATTING_PATTERN, "", text)

    if not ProOutputFormatterConfig.use_utf8:
        for char, alternative in UTF8_ALTERNATIVES.items():
            output = output.replace(char, alternative)
        output = output.encode("ascii", "ignore").decode()

    if not sys.stdout.isatty():
        output = re.sub(LINK_START_PATTERN, "", output)
        output = output.replace(LINK_END, "")

    return output


# We can't rely on textwrap because of the real_len function
# Textwrap is using a magic regex instead
def wrap_text(text: str, max_width: int) -> List[str]:
    if real_len(text) <= max_width:
        return [text]

    words = text.split()
    wrapped_lines = []
    current_line = ""

    for word in words:
        if real_len(current_line) + real_len(word) >= max_width:
            wrapped_lines.append(current_line.strip())
            current_line = word
        else:
            current_line += " " + word

    if current_line:
        wrapped_lines.append(current_line.strip())

    return wrapped_lines


class ProOutputFormatter(abc.ABC):
    @abc.abstractmethod
    def to_string(self, line_length: Optional[int] = None) -> str:
        pass

    def __str__(self):
        return self.to_string()


class Table(ProOutputFormatter):
    SEPARATOR = " " * 2

    def __init__(
        self,
        headers: Optional[List[str]] = None,
        rows: Optional[List[List[str]]] = None,
        alignment: Optional[List[ContentAlignment]] = None,
    ):
        self.headers = headers if headers is not None else []
        self.rows = rows if rows is not None else []
        self.column_sizes = self._get_column_sizes()
        self.alignment = (
            alignment
            if alignment is not None
            else [ContentAlignment.LEFT] * len(self.column_sizes)
        )
        if len(self.alignment) != len(self.column_sizes):
            raise ValueError(
                "'alignment' list should have length {}".format(
                    len(self.column_sizes)
                )
            )
        self.last_column_size = self.column_sizes[-1]

    @staticmethod
    def ljust(string: str, total_length: int) -> str:
        str_length = real_len(string)
        if str_length >= total_length:
            return string
        return string + " " * (total_length - str_length)

    @staticmethod
    def rjust(string: str, total_length: int) -> str:
        str_length = real_len(string)
        if str_length >= total_length:
            return string
        return " " * (total_length - str_length) + string

    def _get_column_sizes(self) -> List[int]:
        if not self.headers and not self.rows:
            raise ValueError(
                "Empty table not supported. Please provide headers or rows."
            )

        if self.rows and any(len(item) == 0 for item in self.rows):
            raise ValueError(
                "Empty row not supported. Please provide content for each row."
            )

        all_content = []
        if self.headers:
            all_content.append(self.headers)
        if self.rows:
            all_content.extend(self.rows)

        expected_length = len(all_content[0])
        if not all(len(item) == expected_length for item in all_content):
            raise ValueError(
                "Mixed lengths in table content. "
                "Please provide headers / rows of the same length."
            )

        column_sizes = []
        for i in range(len(all_content[0])):
            column_sizes.append(
                max(real_len(str(item[i])) for item in all_content)
            )

        return column_sizes

    def to_string(self, line_length: Optional[int] = None) -> str:
        if line_length is None:
            line_length = _get_default_length()

        rows = self.rows
        if self._get_line_length() > line_length:
            rows = self.wrap_last_column(line_length)
        output = ""
        if self.headers:
            output += (
                TxtColor.BOLD
                + self._fill_row(self.headers)
                + TxtColor.ENDC
                + "\n"
            )
        for row in rows:
            output += self._fill_row(row)
            output += "\n"

        return process_formatter_config(output)

    def _get_line_length(self) -> int:
        return sum(self.column_sizes) + (len(self.column_sizes) - 1) * len(
            self.SEPARATOR
        )

    def wrap_last_column(self, max_length: int) -> List[List[str]]:
        self.last_column_size = max_length - (
            sum(self.column_sizes[:-1])
            + (len(self.column_sizes) - 1) * len(self.SEPARATOR)
        )
        new_rows = []
        for row in self.rows:
            if len(row[-1]) <= self.last_column_size:
                new_rows.append(row)
            else:
                wrapped_last_column = wrap_text(row[-1], self.last_column_size)
                new_rows.append(row[:-1] + [wrapped_last_column[0]])
                for extra_line in wrapped_last_column[1:]:
                    new_row = [" "] * (len(self.column_sizes) - 1) + [
                        extra_line
                    ]
                    new_rows.append(new_row)
        return new_rows

    def _fill_row(self, row: List[str]) -> str:
        output = ""
        for i in range(len(row) - 1):
            if self.alignment[i] == ContentAlignment.LEFT:
                output += (
                    self.ljust(row[i], self.column_sizes[i]) + self.SEPARATOR
                )
            elif self.alignment[i] == ContentAlignment.RIGHT:
                output += (
                    self.rjust(row[i], self.column_sizes[i]) + self.SEPARATOR
                )
        if self.alignment[-1] == ContentAlignment.LEFT:
            output += row[-1]
        elif self.alignment[-1] == ContentAlignment.RIGHT:
            output += self.rjust(row[-1], self.last_column_size)
        return output


class Block(ProOutputFormatter):
    INDENT_SIZE = 4
    INDENT_CHAR = " "

    def __init__(
        self,
        title: Optional[str] = None,
        content: Optional[List[Any]] = None,
    ):
        self.title = title
        self.content = content if content is not None else []

    def to_string(self, line_length: Optional[int] = None) -> str:
        if line_length is None:
            line_length = _get_default_length()

        line_length -= self.INDENT_SIZE

        output = ""

        if self.title:
            output += (
                TxtColor.BOLD
                + TxtColor.DISABLEGREY
                + self.title
                + TxtColor.ENDC
                + "\n"
            )

        for item in self.content:
            if isinstance(item, ProOutputFormatter):
                item_str = item.to_string(line_length=line_length)
            else:
                item_str = "\n".join(wrap_text(str(item), line_length)) + "\n"

            output += textwrap.indent(
                item_str, self.INDENT_CHAR * self.INDENT_SIZE
            )

        return process_formatter_config(output)



import logging

from uaclient import apt_news, config, contract, exceptions, messages, util
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.parser import HelpCategory
from uaclient.timer.update_messaging import refresh_motd, update_motd_messages

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def _action_refresh_config(args, cfg: config.UAConfig):
    try:
        cfg.process_config()
    except RuntimeError as exc:
        LOG.exception(exc)
        raise exceptions.RefreshConfigFailure()
    print(messages.REFRESH_CONFIG_SUCCESS)


@cli_util.assert_attached()
def _action_refresh_contract(_args, cfg: config.UAConfig):
    try:
        contract.refresh(cfg)
    except exceptions.ConnectivityError:
        raise exceptions.RefreshContractFailure()
    print(messages.REFRESH_CONTRACT_SUCCESS)


def _action_refresh_messages(_args, cfg: config.UAConfig):
    # Not performing any exception handling here since both of these
    # functions should raise UbuntuProError exceptions, which are
    # covered by the main_error_handler decorator
    try:
        update_motd_messages(cfg)
        refresh_motd()
        if cfg.apt_news:
            apt_news.update_apt_news(cfg)
    except Exception as exc:
        LOG.exception(exc)
        raise exceptions.RefreshMessagesFailure()
    else:
        print(messages.REFRESH_MESSAGES_SUCCESS)


@cli_util.assert_root
@cli_util.assert_lock_file("pro refresh")
def action_refresh(args, *, cfg: config.UAConfig, **kwargs):
    if args.target is None or args.target == "config":
        _action_refresh_config(args, cfg)

    if args.target is None or args.target == "contract":
        _action_refresh_contract(args, cfg)

    if args.target is None or args.target == "messages":
        _action_refresh_messages(args, cfg)

    return 0


refresh_command = ProCommand(
    "refresh",
    help=messages.CLI_ROOT_REFRESH,
    description=messages.CLI_REFRESH_DESC,
    action=action_refresh,
    preserve_description=True,
    help_category=HelpCategory.OTHER,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "target",
                    help=messages.CLI_REFRESH_TARGET,
                    nargs="?",
                    choices=["contract", "config", "messages"],
                    default=None,
                )
            ]
        )
    ],
)



import datetime

import mock
import pytest

from uaclient import messages
from uaclient.api.u.pro.packages.updates.v1 import (
    PackageUpdatesResult,
    UpdateSummary,
)
from uaclient.api.u.pro.status.is_attached.v1 import ContractExpiryStatus
from uaclient.files import notices
from uaclient.timer.update_messaging import (
    update_contract_expiry,
    update_motd_messages,
)

M_PATH = "uaclient.timer.update_messaging."


class TestGetContractExpiryStatus:
    @pytest.mark.parametrize(
        "expiry,is_updated",
        (
            (
                datetime.datetime(
                    2040,
                    5,
                    8,
                    19,
                    2,
                    26,
                    tzinfo=datetime.timezone.utc,
                ),
                False,
            ),
            (
                datetime.datetime(
                    2042,
                    5,
                    8,
                    19,
                    2,
                    26,
                    tzinfo=datetime.timezone.utc,
                ),
                True,
            ),
        ),
    )
    @mock.patch("uaclient.files.MachineTokenFile.write")
    @mock.patch(M_PATH + "contract.UAContractClient.get_contract_machine")
    def test_update_contract_expiry(
        self,
        m_get_contract_machine,
        m_machine_token_write,
        expiry,
        is_updated,
        FakeConfig,
        fake_machine_token_file,
    ):
        m_get_contract_machine.return_value = {
            "machineTokenInfo": {"contractInfo": {"effectiveTo": expiry}}
        }
        fake_machine_token_file.attached = True
        update_contract_expiry(FakeConfig())
        if is_updated:
            assert 1 == fake_machine_token_file.write_calls
        else:
            assert 0 == fake_machine_token_file.write_calls


class TestUpdateMotdMessages:
    @pytest.mark.parametrize(
        [
            "is_attached_side_effect",
            "is_current_series_active_esm",
            "infra_enabled",
            "is_current_series_lts",
            "apps_enabled",
            "updates",
            "expected",
            "update_contract_expiry_calls",
            "notices_remove_calls",
            "ensure_file_absent_calls",
            "write_file_calls",
            "notices_add_calls",
        ],
        [
            (
                # not attached
                [
                    mock.MagicMock(
                        is_attached=False,
                        contract_status=None,
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=False,
                    )
                ],
                False,
                False,
                False,
                False,
                None,
                False,
                [],
                [],
                [],
                [],
                [],
            ),
            (
                # somehow attached but none contract status
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.NONE.value,
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=False,
                    )
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
                [mock.call(mock.ANY)],
                [],
                [],
            ),
            (
                # active contract
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.ACTIVE.value,
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=True,
                    )
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
                [mock.call(mock.ANY)],
                [],
                [],
            ),
            (
                # expiring soon contract, updated to be active
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.ACTIVE_EXPIRED_SOON.value,  # noqa
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.ACTIVE.value,
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
                [mock.call(mock.ANY)],
                [],
                [],
            ),
            (
                # expired grace period contract, updated to be active
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED_GRACE_PERIOD.value,  # noqa
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.ACTIVE.value,
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
                [mock.call(mock.ANY)],
                [],
                [],
            ),
            (
                # expired contract, updated to be active
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.ACTIVE.value,
                        contract_remaining_days=0,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
                [mock.call(mock.ANY)],
                [],
                [],
            ),
            (
                # expiring soon for real
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.ACTIVE_EXPIRED_SOON.value,  # noqa
                        contract_remaining_days=3,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.ACTIVE_EXPIRED_SOON.value,  # noqa
                        contract_remaining_days=3,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
                [],
                [
                    mock.call(
                        mock.ANY,
                        messages.CONTRACT_EXPIRES_SOON.pluralize(3).format(
                            remaining_days=3
                        )
                        + "\n\n",
                    )
                ],
                [],
            ),
            (
                # expired grace period for real
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED_GRACE_PERIOD.value,  # noqa
                        contract_remaining_days=-3,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED_GRACE_PERIOD.value,  # noqa
                        contract_remaining_days=-3,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
                [],
                [
                    mock.call(
                        mock.ANY,
                        messages.CONTRACT_EXPIRED_GRACE_PERIOD.pluralize(
                            11
                        ).format(remaining_days=11, expired_date="21 Dec 2012")
                        + "\n\n",
                    )
                ],
                [],
            ),
            (
                # expired, eol release, esm-infra not enabled
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                True,
                False,
                False,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [],
                [],
                [mock.call(mock.ANY, messages.CONTRACT_EXPIRED + "\n\n")],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
            ),
            (
                # expired, lts release, esm-apps not enabled
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                True,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [],
                [],
                [mock.call(mock.ANY, messages.CONTRACT_EXPIRED + "\n\n")],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
            ),
            (
                # expired, interim release
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                False,
                False,
                None,
                True,
                [mock.call(mock.ANY)],
                [],
                [],
                [mock.call(mock.ANY, messages.CONTRACT_EXPIRED + "\n\n")],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
            ),
            (
                # expired, eol release, esm-infra enabled
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                True,
                True,
                False,
                False,
                PackageUpdatesResult(UpdateSummary(0, 0, 4, 0, 0), []),
                True,
                [mock.call(mock.ANY)],
                [],
                [],
                [
                    mock.call(
                        mock.ANY,
                        messages.CONTRACT_EXPIRED_WITH_PKGS.pluralize(
                            4
                        ).format(service="esm-infra", pkg_num=4)
                        + "\n\n",
                    )
                ],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
            ),
            (
                # expired, lts release, esm-apps enabled
                [
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                    mock.MagicMock(
                        is_attached=True,
                        contract_status=ContractExpiryStatus.EXPIRED.value,
                        contract_remaining_days=-30,
                        is_attached_and_contract_valid=True,
                    ),
                ],
                False,
                False,
                True,
                True,
                PackageUpdatesResult(UpdateSummary(0, 5, 0, 0, 0), []),
                True,
                [mock.call(mock.ANY)],
                [],
                [],
                [
                    mock.call(
                        mock.ANY,
                        messages.CONTRACT_EXPIRED_WITH_PKGS.pluralize(
                            5
                        ).format(service="esm-apps", pkg_num=5)
                        + "\n\n",
                    )
                ],
                [mock.call(notices.Notice.CONTRACT_EXPIRED)],
            ),
        ],
    )
    @mock.patch(M_PATH + "_enabled_services")
    @mock.patch(M_PATH + "api_u_pro_packages_updates_v1")
    @mock.patch(M_PATH + "system.is_current_series_lts")
    @mock.patch(M_PATH + "system.is_current_series_active_esm")
    @mock.patch(M_PATH + "notices.add")
    @mock.patch(M_PATH + "system.write_file")
    @mock.patch(M_PATH + "system.ensure_file_absent")
    @mock.patch(M_PATH + "notices.remove")
    @mock.patch(M_PATH + "update_contract_expiry")
    @mock.patch(M_PATH + "_is_attached")
    def test_update_motd_messages(
        self,
        m_is_attached,
        m_update_contract_expiry,
        m_notices_remove,
        m_ensure_file_absent,
        m_write_file,
        m_notices_add,
        m_is_current_series_active_esm,
        m_is_current_series_lts,
        m_api_updates_v1,
        m_enabled_services,
        is_attached_side_effect,
        is_current_series_active_esm,
        infra_enabled,
        is_current_series_lts,
        apps_enabled,
        updates,
        expected,
        update_contract_expiry_calls,
        notices_remove_calls,
        ensure_file_absent_calls,
        write_file_calls,
        notices_add_calls,
        FakeConfig,
        fake_machine_token_file,
    ):
        enabled_services = []
        if infra_enabled:
            m_infra = mock.MagicMock()
            type(m_infra).name = mock.PropertyMock(return_value="esm-infra")
            enabled_services.append(m_infra)

        if apps_enabled:
            m_apps = mock.MagicMock()
            type(m_apps).name = mock.PropertyMock(return_value="esm-apps")
            enabled_services.append(m_apps)

        m_enabled_services.return_value = mock.MagicMock(
            enabled_services=enabled_services
        )
        m_is_attached.side_effect = is_attached_side_effect
        m_is_current_series_active_esm.return_value = (
            is_current_series_active_esm
        )
        m_is_current_series_lts.return_value = is_current_series_lts
        m_api_updates_v1.return_value = updates
        print(updates)

        fake_machine_token_file._contract_expiry_datetime = datetime.datetime(
            2012, 12, 21
        )
        assert expected == update_motd_messages(FakeConfig())

        assert (
            update_contract_expiry_calls
            == m_update_contract_expiry.call_args_list
        )
        assert notices_remove_calls == m_notices_remove.call_args_list
        assert ensure_file_absent_calls == m_ensure_file_absent.call_args_list
        assert write_file_calls == m_write_file.call_args_list
        assert notices_add_calls == m_notices_add.call_args_list



import time

from uaclient import actions, config, event_logger, messages, status, util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.parser import HelpCategory

event = event_logger.get_event_logger()


def action_status(args, *, cfg: config.UAConfig, **kwargs):
    if not cfg:
        cfg = config.UAConfig()
    show_all = args.all if args else False
    token = args.simulate_with_token if args else None
    active_value = status.UserFacingConfigStatus.ACTIVE.value
    status_dict, ret = actions.status(
        cfg, simulate_with_token=token, show_all=show_all
    )
    config_active = bool(status_dict["execution_status"] == active_value)

    if args and args.wait and config_active:
        while status_dict["execution_status"] == active_value:
            event.info(".", end="")
            time.sleep(1)
            status_dict, ret = actions.status(
                cfg,
                simulate_with_token=token,
                show_all=show_all,
            )
        event.info("")

    event.set_output_content(status_dict)
    output = status.format_tabular(status_dict, show_all=show_all)
    event.info(util.handle_unicode_characters(output))
    event.process_events()
    return ret


status_command = ProCommand(
    "status",
    help=messages.CLI_ROOT_STATUS,
    description=messages.CLI_STATUS_DESC,
    action=action_status,
    preserve_description=True,
    help_category=HelpCategory.QUICKSTART,
    help_position=1,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "--wait",
                    help=messages.CLI_STATUS_WAIT,
                    action="store_true",
                ),
                ProArgument(
                    "--format",
                    help=messages.CLI_FORMAT_DESC.format(default="tabular"),
                    action="store",
                    choices=["tabular", "json", "yaml"],
                    default="tabular",
                ),
                ProArgument(
                    "--simulate-with-token",
                    help=messages.CLI_STATUS_SIMULATE_WITH_TOKEN,
                    metavar="TOKEN",
                    action="store",
                ),
                ProArgument(
                    "--all", help=messages.CLI_STATUS_ALL, action="store_true"
                ),
            ]
        )
    ],
)



import mock
import pytest

from uaclient import messages, util
from uaclient.entitlements.entitlement_status import ApplicabilityStatus
from uaclient.entitlements.realtime import (
    GenericRealtime,
    IntelIotgRealtime,
    RaspberryPiRealtime,
)
from uaclient.system import CpuInfo

M_PATH = "uaclient.entitlements.realtime."


class TestGenericVariant:
    @pytest.mark.parametrize(
        [
            "current_kernel",
            "expected_pre_enable_messages",
        ],
        [
            (mock.MagicMock(flavor="generic"), None),
            (
                mock.MagicMock(flavor="nonstandard"),
                [
                    (
                        util.prompt_for_confirmation,
                        {
                            "msg": messages.KERNEL_FLAVOR_CHANGE_WARNING_PROMPT.format(  # noqa: E501
                                service="realtime-kernel",
                                variant="generic",
                                base_flavor="generic",
                                current_flavor="nonstandard",
                            )
                        },
                    )
                ],
            ),
        ],
    )
    @mock.patch(M_PATH + "system.get_kernel_info")
    @mock.patch(
        M_PATH + "RealtimeKernelEntitlement.messaging",
        new_callable=mock.PropertyMock,
        return_value={},
    )
    def test_messaging_on_wrong_kernel(
        self,
        _m_super_messaging,
        m_get_kernel_info,
        current_kernel,
        expected_pre_enable_messages,
        FakeConfig,
    ):
        m_get_kernel_info.return_value = current_kernel
        assert expected_pre_enable_messages == GenericRealtime(
            FakeConfig()
        ).messaging.get("pre_enable")


class TestRaspiVariant:
    @pytest.mark.parametrize(
        [
            "load_file_side_effect",
            "expected_result",
        ],
        [
            (["Raspberry Pi 5 Model B Rev 1.0"], True),
            (["Raspberry Pi 4 Model B Rev 1.1"], True),
            (["Raspberry Pi 3 Model B Plus Rev 1.3"], False),
            ([FileNotFoundError()], False),
        ],
    )
    @mock.patch(M_PATH + "system.load_file")
    def test_variant_auto_select(
        self,
        m_load_file,
        load_file_side_effect,
        expected_result,
        FakeConfig,
    ):
        m_load_file.side_effect = load_file_side_effect
        assert (
            expected_result
            == RaspberryPiRealtime(FakeConfig()).variant_auto_select()
        )


class TestIntelIOTGVariannt:
    @pytest.mark.parametrize(
        "cpu_info,expected_status,expected_msg",
        (
            (
                CpuInfo(
                    cpuinfo_vendor_id="test",
                    cpuinfo_model=None,
                    cpuinfo_stepping=None,
                    cpuinfo_cpu=None,
                    cpuinfo_cpu_family=None,
                    cpuinfo_cpu_implementer=None,
                    cpuinfo_cpu_architecture=None,
                    cpuinfo_cpu_part=None,
                    cpuinfo_cpu_revision=None,
                    cpuinfo_cpu_variant=None,
                    cpuinfo_model_name=None,
                    sysinfo_model=None,
                    sysinfo_type=None,
                    sys_firmware_devicetree_base_model=None,
                ),
                ApplicabilityStatus.INAPPLICABLE,
                messages.INAPPLICABLE_VENDOR_NAME.format(
                    title=IntelIotgRealtime.title,
                    vendor="test",
                    supported_vendors="intel",
                ),
            ),
            (
                CpuInfo(
                    cpuinfo_vendor_id="GenuineIntel",
                    cpuinfo_model=None,
                    cpuinfo_stepping=None,
                    cpuinfo_cpu=None,
                    cpuinfo_cpu_family=None,
                    cpuinfo_cpu_implementer=None,
                    cpuinfo_cpu_architecture=None,
                    cpuinfo_cpu_part=None,
                    cpuinfo_cpu_revision=None,
                    cpuinfo_cpu_variant=None,
                    cpuinfo_model_name=None,
                    sysinfo_model=None,
                    sysinfo_type=None,
                    sys_firmware_devicetree_base_model=None,
                ),
                ApplicabilityStatus.APPLICABLE,
                None,
            ),
        ),
    )
    @mock.patch("uaclient.system.get_kernel_info")
    @mock.patch("uaclient.system.get_cpu_info")
    def test_applicability_status(
        self,
        m_get_cpu_info,
        _m_get_kernel_info,
        cpu_info,
        expected_status,
        expected_msg,
        FakeConfig,
    ):
        m_get_cpu_info.return_value = cpu_info
        ent = IntelIotgRealtime(FakeConfig())
        with mock.patch.object(
            ent, "_base_entitlement_cfg"
        ) as m_entitlement_cfg:
            m_entitlement_cfg.return_value = {
                "entitlement": {
                    "affordances": {
                        "platformChecks": {
                            "cpu_vendor_ids": ["intel"],
                        }
                    }
                }
            }
            actual_ret = ent.applicability_status()
            assert (
                expected_status,
                expected_msg,
            ) == actual_ret
